import numpy
import os.path

from amuse.test.amusetest import TestWithMPI
from amuse.community.gadget2.interface import Gadget2Interface, Gadget2
from amuse.ext.evrard_test import MakeEvrardTest, new_evrard_gas_sphere, body_centered_grid_unit_cube
from amuse.ext.spherical_model import new_uniform_spherical_particle_distribution
from amuse.support.exceptions import AmuseException

from amuse.units import nbody_system
from amuse.units import generic_unit_converter
from amuse.units import generic_unit_system
from amuse.units import units, constants
from amuse.datamodel import Particles, Grid, ParticlesSuperset, new_regular_grid
from amuse.rfi import channel
from amuse.io import read_set_from_file, write_set_to_file
from amuse.ic.plummer import new_plummer_model
from amuse.ic.gasplummer import new_plummer_gas_model
from amuse.ext.cosmo import convert_comoving_to_physical

try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
    from amuse.plot import plot, xlabel, ylabel
except ImportError:
    HAS_MATPLOTLIB = False

default_options = dict(number_of_workers=1)
# default_options = dict(number_of_workers=2, redirection="none")

# ... but never use (number_of_workers>1) for tests with only a few particles:
few_particles_default_options = dict()
# few_particles_default_options = dict(redirection="none")

testing_isotherm_no_gravity = False


class TestGadget2Interface(TestWithMPI):

    def test1(self):
        instance = Gadget2Interface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        instance = Gadget2Interface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([1.989000e43, 0], list(instance.get_unit_mass().values()))
        self.assertEqual([3.085678e21, 0], list(instance.get_unit_length().values()))
        self.assertEqual([3.085678e16, 0], list(instance.get_unit_time().values()))
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test3(self):
        instance = Gadget2Interface(**few_particles_default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([1, 0], list(instance.new_dm_particle(0.01,  1, 0, 0,  0, 1, 0).values()))
        self.assertEqual([2, 0], list(instance.new_dm_particle(0.01, -1, 0, 0,  0, -1, 0).values()))
        self.assertEqual(-1, instance.get_index_of_first_particle()['__result'])
        self.assertEqual(0, instance.commit_particles())
        self.assertEqual([1, 0], list(instance.get_index_of_first_particle().values()))
        self.assertEqual([2, 1], list(instance.get_index_of_next_particle(1).values()))
        self.assertEqual(-1, instance.get_index_of_next_particle(2)['__result'])
        self.assertEqual(0, instance.evolve_model(0.01))
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test3a(self):
        instance = Gadget2Interface(**few_particles_default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        # cgs
        instance.set_unit_mass(0.001)
        instance.set_unit_length(1.0)
        instance.set_unit_time(122404.614048)
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([1, 0], list(instance.new_dm_particle(1.,  1, 0, 0,  0, 0, 0).values()))
        self.assertEqual([2, 0], list(instance.new_dm_particle(1., -1, 0, 0,  0, 0, 0).values()))
        self.assertEqual(0, instance.commit_particles())
        if testing_isotherm_no_gravity:
            self.assertAlmostEqual(0.000, instance.get_potential(1)['Potential'], places=1)
        else:
            self.assertAlmostEqual(-0.500, instance.get_potential(1)['Potential'], places=1)
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test4(self):
        instance = Gadget2Interface(**few_particles_default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([1, 0], list(instance.new_dm_particle(0.01,  1, 0, 0,  0, 1, 0).values()))
        self.assertEqual([2, 0], list(instance.new_dm_particle(0.02, -1, 0, 0,  0, -1, 0).values()))
        self.assertEqual(-3, instance.get_mass(1)['__result'])
        self.assertEqual(0, instance.commit_particles())

        mass, result = instance.get_mass(1)
        self.assertAlmostEqual(0.01, mass)
        self.assertEqual(0, result)
        mass, result = instance.get_mass(2)
        self.assertAlmostEqual(0.02, mass)
        self.assertEqual(0, result)
        self.assertEqual(-3, instance.get_mass(3)['__result'])
        for result, expected in zip(instance.get_position(1), [1, 0, 0, 0]):
            self.assertAlmostEqual(result, expected)
        for result, expected in zip(instance.get_position(2), [-1, 0, 0, 0]):
            self.assertAlmostEqual(result, expected)
        for result, expected in zip(instance.get_velocity(1), [0, 1, 0, 0]):
            self.assertAlmostEqual(result, expected)
        for result, expected in zip(instance.get_velocity(2), [0, -1, 0, 0]):
            self.assertAlmostEqual(result, expected)

        self.assertEqual(0, instance.set_state(1, 0.01, 1, 2, 3, 4, 5, 6))
        self.assertEqual([0.01, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0], list(instance.get_state(1).values()))

        self.assertEqual(0, instance.evolve_model(0.01))
        mass, result = instance.get_mass(1)
        self.assertAlmostEqual(0.01, mass)
        self.assertEqual(0, result)
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test5(self):
        instance = Gadget2Interface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        target_number_of_particles = 100
        evrard = MakeEvrardTest(target_number_of_particles)
        mass, x, y, z, vx, vy, vz, u = evrard.new_model()
        number_of_particles = len(mass)
        indices, results = instance.new_sph_particle(mass, x, y, z, vx, vy, vz, u)
        self.assertEqual([0 for i in range(number_of_particles)], list(results))
        self.assertEqual([i+1 for i in range(number_of_particles)], list(indices))
        self.assertEqual(0, instance.commit_particles())
        self.assertEqual(0, instance.evolve_model(0.00005))
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test6(self):
        instance = Gadget2Interface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        target_number_of_particles = 100
        evrard = MakeEvrardTest(target_number_of_particles)
        mass, x, y, z, vx, vy, vz, u = evrard.new_model()
        number_of_particles = len(mass)
        first_half = number_of_particles//2
        indices, results = instance.new_sph_particle(mass[:first_half], x[:first_half], y[:first_half], z[:first_half],
            vx[:first_half], vy[:first_half], vz[:first_half], u[:first_half])
        self.assertEqual([0 for i in range(first_half)], list(results))
        self.assertEqual([i+1 for i in range(first_half)], list(indices))
        self.assertEqual([number_of_particles/2+1, 0], list(instance.new_dm_particle(0.01,  1, 0, 0,  0, 1, 0).values()))
        self.assertEqual([number_of_particles/2+2, 0], list(instance.new_dm_particle(0.02, -1, 0, 0,  0, -1, 0).values()))
        indices, results = instance.new_sph_particle(mass[first_half:], x[first_half:], y[first_half:], z[first_half:],
            vx[first_half:], vy[first_half:], vz[first_half:], u[first_half:])
        self.assertEqual([0 for i in range(number_of_particles-first_half)], list(results))
        self.assertEqual([first_half+i+1+2 for i in range(number_of_particles-first_half)], list(indices))
        self.assertEqual(0, instance.commit_particles())

        mass_list = [x for sublist in [mass[:first_half], [0.01, 0.02], mass[first_half:]] for x in sublist]
        first_index, result = list(instance.get_index_of_first_particle().values())
        self.assertEqual([1, 0], [first_index, result])
        for i in range(first_index, number_of_particles+2):
            index, result = instance.get_index_of_next_particle(i)
            self.assertEqual([i+1, 0 if (i < number_of_particles+1) else 1], [index, result])
            mass, result = instance.get_mass(index)
            self.assertEqual(0, result)
            self.assertAlmostEqual(mass_list[i], mass)

        self.assertEqual(0, instance.evolve_model(0.0001))
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test7(self):
        instance = Gadget2Interface(**default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        target_number_of_particles = 100
        evrard = MakeEvrardTest(target_number_of_particles)
        mass, x, y, z, vx, vy, vz, u = evrard.new_model()
        number_of_particles = len(mass)
        indices, results = instance.new_sph_particle(mass, x, y, z, vx, vy, vz, u)
        self.assertEqual([0 for i in range(number_of_particles)], list(results))
        self.assertEqual([i+1 for i in range(number_of_particles)], list(indices))
        self.assertEqual([number_of_particles+1, 0], list(instance.new_dm_particle(0.01,  1, 0, 0,  0, 1, 0).values()))
        self.assertEqual([number_of_particles+2, 0], list(instance.new_dm_particle(0.02, -1, 0, 0,  0, -1, 0).values()))
        self.assertEqual(0, instance.commit_particles())
        self.assertEqual(0, instance.evolve_model(0.001))
        self.assertEqual(0, instance.delete_particle(number_of_particles-1))
        self.assertEqual(0, instance.delete_particle(number_of_particles+1))
        self.assertEqual(-3, instance.delete_particle(number_of_particles-1))
        indices, results = instance.new_sph_particle(mass, x, y, z, vx, vy, vz, u)
        self.assertEqual([2*number_of_particles+3, 0], list(instance.new_dm_particle(0.02, -1, 0, 0,  0, -1, 0).values()))
        self.assertEqual(0, instance.recommit_particles())
        mass_list = [x for sublist in [mass[:number_of_particles-2], [mass[number_of_particles-1], 0.02],
            mass, [0.02]] for x in sublist]
        index_list = [x for sublist in [list(range(1, number_of_particles-1)), [number_of_particles],
            list(range(number_of_particles+2, 2*number_of_particles+4))] for x in sublist]
        index, result = list(instance.get_index_of_first_particle().values())
        self.assertEqual([index_list[0], 0], [index, result])
        for i in range(1, 2*number_of_particles+1):
            index, result = instance.get_index_of_next_particle(index)
            self.assertEqual([index_list[i], 0 if (i < 2*number_of_particles) else 1], [index, result])
            mass, result = instance.get_mass(index)
            self.assertEqual(0, result)
            self.assertAlmostEqual(mass_list[i], mass)

        self.assertEqual(0, instance.evolve_model(0.001))
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test8(self):
        instance = Gadget2Interface(**few_particles_default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())
        for i in range(1, 10):
            self.assertEqual([i, 0], list(instance.new_dm_particle(i*0.01,  i, -i, 0,  -i*10, i*10, 0).values()))
        self.assertEqual(0, instance.commit_particles())

        for i in range(1, 10):
            mass, result = instance.get_mass(i)
            self.assertAlmostEqual(0.01*i, mass)
            self.assertEqual(0, result)

            x, y, z, result = instance.get_position(i)
            self.assertAlmostEqual(i, x)
            self.assertAlmostEqual(-i, y)
            self.assertAlmostEqual(0, z)
            self.assertEqual(0, result)

            vx, vy, vz, result = instance.get_velocity(i)
            self.assertAlmostEqual(-i*10, vx)
            self.assertAlmostEqual(i*10, vy)
            self.assertAlmostEqual(0, vz)
            self.assertEqual(0, result)

            self.assertEqual(0, instance.set_mass(i, i*0.1))
            mass, result = instance.get_mass(i)
            self.assertAlmostEqual(0.1*i, mass)
            self.assertEqual(0, result)

            self.assertEqual(0, instance.set_position(i, 2*i, -2*i, 0))
            x, y, z, result = instance.get_position(i)
            self.assertAlmostEqual(2*i, x)
            self.assertAlmostEqual(-2*i, y)
            self.assertAlmostEqual(0, z)
            self.assertEqual(0, result)

            self.assertEqual(0, instance.set_velocity(i, -i*20, i*20, 0))
            vx, vy, vz, result = instance.get_velocity(i)
            self.assertAlmostEqual(-i*20, vx)
            self.assertAlmostEqual(i*20, vy)
            self.assertAlmostEqual(0, vz)
            self.assertEqual(0, result)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test9(self):
        instance = Gadget2Interface(mode=Gadget2Interface.MODE_PERIODIC_NOGRAVITY,
            **few_particles_default_options)
        instance.initialize_code()
        instance.set_min_size_timestep(1.0)
        instance.set_gadget_output_directory(instance.get_output_directory())

        instance.set_box_size(2.)
        value, error = instance.get_box_size()
        self.assertEqual(error, 0)
        self.assertEqual(value, 2.)

        self.assertEqual(instance.get_periodic_boundaries_flag()['value'], True)
        self.assertEqual(instance.set_periodic_boundaries_flag(True), -2)  # read-only
        self.assertEqual(instance.get_nogravity()['no_gravity_flag'], True)
        self.assertEqual(instance.commit_parameters(), 0)
        ids, err = instance.new_particle(
           [1.0, 1.0, 1.0],
           [0.5, 0.0, 0.0],
           [0.0, -0.5, 0.0],
           [0.0, 0.0, 0.5],
           [-1.0, 0.0, 0.0],
           [0.0, 1.0, 0.0],
           [0.0, 0.0, -1.0])
        instance.commit_particles()
        m, x, y, z, vx, vy, vz, err = instance.get_state(ids)
        self.assertAlmostEqual(x, [0.5, 0., 0.], places=6)
        self.assertAlmostEqual(y, [0., -0.5, 0.], places=6)
        self.assertAlmostEqual(z, [0., 0., 0.5], places=6)
        instance.evolve_model(0.1)
        m, x, y, z, vx, vy, vz, err = instance.get_state(ids)
        self.assertAlmostEqual(x, [0.4, 0., 0.], places=6)
        self.assertAlmostEqual(y, [0., -0.4, 0.], places=6)
        self.assertAlmostEqual(z, [0., 0., 0.4], places=6)

        instance.evolve_model(1.0)
        m, x, y, z, vx, vy, vz, err = instance.get_state(ids)
        self.assertAlmostEqual(x, [-0.5, 0., 0.], places=6)
        self.assertAlmostEqual(y, [0., 0.5, 0.], places=6)
        self.assertAlmostEqual(z, [0., 0., -0.5], places=6)
        instance.cleanup_code()
        instance.stop()

        instance = Gadget2Interface(mode=Gadget2Interface.MODE_NORMAL,
            **few_particles_default_options)  # MODE_NORMAL is default: non-periodic
        instance.initialize_code()
        instance.set_gadget_output_directory(instance.get_output_directory())

        self.assertEqual(instance.get_periodic_boundaries_flag()['value'], False)  # default, has to be changed for periodic runs
        self.assertEqual(instance.commit_parameters(), 0)
        self.assertEqual(instance.set_periodic_boundaries_flag(True), -2)  # read-only
        instance.stop()

    def test10(self):
        instance = Gadget2Interface(**few_particles_default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_gadget_output_directory(instance.get_output_directory()))
        self.assertEqual(0, instance.commit_parameters())

        indices_and_value = []
        for i in range(10):
            index, error = instance.new_sph_particle(0.1, i, 0, 0, 0, 0, 0, 0.0)
            self.assertEqual(0, error)
            indices_and_value.append((index, i))

        instance.commit_particles()
        for i, value in indices_and_value:
            x, y, z, error = instance.get_position(i)
            self.assertEqual(0, error)
            self.assertEqual(x, value)
            self.assertEqual(y, 0)
            self.assertEqual(z, 0)

        self.assertEqual(0, error)
        for i in range(10, 20):
            index, error = instance.new_sph_particle(0.1, i, 0, 0, 0, 0, 0, 0.0)
            self.assertEqual(0, error)
            indices_and_value.append((index, i))

        error = instance.recommit_particles()
        self.assertEqual(0, error)

        for i, value in indices_and_value:
            x, y, z, error = instance.get_position(i)
            self.assertEqual(0, error)
            self.assertEqual(x, value)
            self.assertEqual(y, 0)
            self.assertEqual(z, 0)

        instance.stop()


class TestGadget2(TestWithMPI):
    classname = "<class 'amuse.community.gadget2.interface.Gadget2'>"

    UnitLength = 3.085678e21 | units.cm     # ~ 1.0 kpc
    UnitMass = 1.989e43 | units.g           # 1.0e10 solar masses
    UnitVelocity = 1e5 | units.cm / units.s  # 1 km/sec
    default_converter = generic_unit_converter.ConvertBetweenGenericAndSiUnits(UnitLength, UnitMass, UnitVelocity)
    default_convert_nbody = nbody_system.nbody_to_si(UnitLength, UnitMass)

    three_particles_IC = Particles(3)
    three_particles_IC.position = [[0.5, 0.0, 0.0], [0.0, -0.5, 0.0], [0.0, 0.0, 0.5]] | units.kpc
    three_particles_IC.velocity = [[-1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0]] | units.km / units.s
    three_particles_IC.mass = 1.0e10 | units.MSun

    def test1(self):
        print("Testing Gadget initialization")
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        self.assertTrue(
            os.path.join("__amuse_code_output", "gadget2")
            in str(instance.parameters.gadget_output_directory)
        )
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Testing Gadget parameters")
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        self.assertAlmostEqual(instance.parameters.epsilon_squared, (0.01 | units.kpc)**2)
        self.assertAlmostRelativeEquals(instance.parameters.code_mass_unit, self.UnitMass, 7)
        self.assertAlmostRelativeEquals(instance.parameters.code_time_unit, 3.085678e16 | units.s, 7)
        self.assertAlmostRelativeEquals(instance.parameters.code_length_unit, self.UnitLength, 7)
        instance.parameters.epsilon_squared = 0.01 | units.kpc**2
        instance.parameters.opening_angle = 0.4
        instance.commit_parameters()
        self.assertAlmostEqual(instance.parameters.epsilon_squared, 0.01 | units.kpc**2)
        self.assertAlmostEqual(instance.parameters.opening_angle, 0.4)
        self.assertAlmostRelativeEquals(instance.parameters.code_mass_unit, 1.989e43 | units.g, 7)
        self.assertAlmostRelativeEquals(instance.parameters.code_time_unit, 3.085678e16 | units.s, 7)
        instance.stop()

    def test3(self):
        print("Testing Gadget, 2 nbody particles")
        # No default_options since this test fails for (number_of_workers > 1):
        instance = Gadget2(self.default_converter, **few_particles_default_options)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.01 | units.kpc**2

        dark = Particles(2)
        dark.mass = [0.4, 0.4] | generic_unit_system.mass
        dark.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.kpc
        dark.velocity = [[100.0, 100.0, 100.0], [1.0, 1.0, 1.0]] | units.km / units.s
        dark.radius = [0.0, 0.0] | units.RSun

        instance.dm_particles.add_particles(dark)
        instance.evolve_model(1.0 | units.Myr)
        self.assertAlmostEqual(instance.model_time, 1.0 | units.Myr, 3)
        instance.stop()

    def test4(self):
        print("Testing Gadget, 100 nbody particles")
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        convert_nbody = nbody_system.nbody_to_si(1.0e9 | units.MSun, 1.0 | units.kpc)
        dark = new_plummer_model(100, convert_nbody)
        instance.dm_particles.add_particles(dark)
        instance.evolve_model(1.0 | units.Myr)
        copied = instance.dm_particles.copy()
        from_copy_to_code = copied.new_channel_to(instance.dm_particles)
        copied.mass *= 2
        from_copy_to_code.copy_attributes(["mass", "x", "y", "z", "vx", "vy", "vz"])
        instance.stop()

    def test5(self):
        print("Test 5: testing SPH particles")
        target_number_sph_particles = 100
        gas = new_evrard_gas_sphere(target_number_sph_particles, self.default_convert_nbody, seed=1234)
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        instance.gas_particles.add_particles(gas)
        instance.evolve_model(0.0001 | generic_unit_system.time)
        instance.stop()

    def test6(self):
        print("Test 6: testing dark matter + SPH particles")
        target_number_sph_particles = 100
        gas = new_evrard_gas_sphere(target_number_sph_particles, self.default_convert_nbody, seed=1234)

        dark = Particles(2)
        dark.mass = [0.4, 0.4] | generic_unit_system.mass
        dark.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.kpc
        dark.velocity = [[100.0, 100.0, 100.0], [1.0, 1.0, 1.0]] | units.km / units.s

        instance = Gadget2(self.default_converter, **default_options)
        instance.dm_particles.add_particles(dark)
        instance.gas_particles.add_particles(gas)
        instance.evolve_model(1.0 | units.Myr)
        instance.stop()

    def test7(self):
        print("Testing more Gadget parameters")
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()

        for par, value in [('gadget_cell_opening_flag', True),
                ('comoving_integration_flag', False),
                ('interpret_kicks_as_feedback', False),
                ('interpret_heat_as_feedback', True)]:
            self.assertTrue(value is getattr(instance.parameters, par))
            setattr(instance.parameters, par, not value)
            self.assertFalse(value is getattr(instance.parameters, par))

        for par, value in [('time_limit_cpu', 36000 | units.s), ('min_gas_temp', 0.0 | units.K)]:
            self.assertEqual(value, getattr(instance.parameters, par))
            setattr(instance.parameters, par, 2 * value)
            self.assertEqual(2 * value, getattr(instance.parameters, par))

        for par, value in [('n_smooth_tol', 0.1), ('n_smooth', 50), ('opening_angle', 0.5),
                ('gadget_cell_opening_constant', 0.005), ('artificial_viscosity_alpha', 0.5),
                ('courant', 0.3), ('type_of_timestep_criterion', 0), ('hubble_parameter', 0.7),
                ('omega_zero', 0.0), ('omega_lambda', 0.0), ('omega_baryon', 0.0),
                ('min_gas_hsmooth_fractional', 0.0), ('timestep_accuracy_parameter', 0.025),
                ('tree_domain_update_frequency', 0.05), ('artificial_viscosity_beta', 1.0)]:
            self.assertEqual(value, getattr(instance.parameters, par))
            setattr(instance.parameters, par, 1)
            self.assertEqual(1, getattr(instance.parameters, par))

        for par, value in [('gas_epsilon', 0.01 | generic_unit_system.length),
                ('begin_time', 0.0 | generic_unit_system.time),
                ('time_max', 100.0 | generic_unit_system.time),
                ('max_size_timestep', 0.01 | generic_unit_system.time),
                ('min_size_timestep', 0.0 | generic_unit_system.time),
                ('periodic_box_size', 1.0 | generic_unit_system.length),
                ('time_between_statistics', 0.1 | generic_unit_system.time),
                ('softening_gas_max_phys', 0.0 | generic_unit_system.length),
                ('softening_halo_max_phys', 0.0 | generic_unit_system.length)]:
            self.assertEqual(instance.unit_converter.to_si(value),
                getattr(instance.parameters, par))
            setattr(instance.parameters, par, 3.0 | value.unit)
            self.assertEqual(instance.unit_converter.to_si(3.0 | value.unit),
                getattr(instance.parameters, par))

        for par, value in [('energy_file', "energy.txt"), ('info_file', "info.txt"),
                ('timings_file', "timings.txt"), ('cpu_file', "cpu.txt")]:
            self.assertEqual(value, getattr(instance.parameters, par))
            setattr(instance.parameters, par, 'test.txt')
            self.assertEqual("test.txt", getattr(instance.parameters, par))

        instance.stop()

    def test8(self):
        print("Testing read-only Gadget parameters")

        def try_set_parameter(par, value, instance):
            setattr(instance.parameters, par, value)

        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        for par, value in [('no_gravity_flag', testing_isotherm_no_gravity),
                            ('isothermal_flag', testing_isotherm_no_gravity),
                            ('eps_is_h_flag',   False),
                            ('periodic_boundaries_flag', False),
                            ('code_mass_unit',     self.default_converter.to_si(generic_unit_system.mass)),
                            ('code_length_unit',   self.default_converter.to_si(generic_unit_system.length)),
                            ('code_time_unit',     self.default_converter.to_si(generic_unit_system.time)),
                            ('code_velocity_unit', self.default_converter.to_si(generic_unit_system.speed)),
                            ('polytropic_index_gamma', (1.0 if testing_isotherm_no_gravity else 5.0/3))]:
            self.assertEqual(value, getattr(instance.parameters, par))
            self.assertRaises(AmuseException, try_set_parameter, par, value, instance,
                expected_message="Could not set value for parameter '"+par+"' of a 'Gadget2' object, "
                    "parameter is read-only")
        instance.stop()

        instance = Gadget2(self.default_converter, mode="nogravity", **default_options)
        instance.initialize_code()
        self.assertEqual(instance.parameters.no_gravity_flag, True)
        self.assertRaises(AmuseException, try_set_parameter, 'no_gravity_flag', False, instance,
            expected_message="Could not set value for parameter 'no_gravity_flag' of a 'Gadget2' object, "
                "parameter is read-only")
        instance.stop()

    def test9(self):
        print("Testing Gadget properties")
        target_number_of_particles = 100
        gas = new_evrard_gas_sphere(target_number_of_particles, self.default_convert_nbody, do_scale=True, seed=1234)
        instance = Gadget2(self.default_converter, **default_options)
        instance.gas_particles.add_particles(gas)
        self.assertEqual(instance.model_time,                        0.0 | units.s)
        if testing_isotherm_no_gravity:
            self.assertAlmostEqual(instance.potential_energy,            0.0 | 1e+50*units.J)
        else:
            self.assertAlmostEqual(instance.potential_energy, -4.27843220393 | 1e+50*units.J)
        self.assertAlmostEqual(instance.kinetic_energy,              0.0 | 1e+49*units.J)
        self.assertAlmostEqual(instance.thermal_energy,    4.27851824913 | 1e+49*units.J)
        self.assertAlmostEqual(instance.total_radius,      3.96592921066 | 1e+19*units.m)
        self.assertAlmostEqual(instance.center_of_mass_position, [0, 0, 0] | 1e+19*units.m)
        self.assertAlmostEqual(instance.center_of_mass_velocity, [0, 0, 0] | units.m/units.s)
        self.assertAlmostEqual(instance.total_mass,                1.989 | 1e+40*units.kg)
        instance.stop()

    def test10(self):
        print("Testing Gadget states")
        target_number_sph_particles = 100
        gas = new_evrard_gas_sphere(target_number_sph_particles, self.default_convert_nbody, seed=1234)
        dark = Particles(2)
        dark.mass = [0.4, 0.4] | generic_unit_system.mass
        dark.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.kpc
        dark.velocity = [[100.0, 100.0, 100.0], [1.0, 1.0, 1.0]] | units.km / units.s

        print("First do everything manually:")
        instance = Gadget2(self.default_converter, **default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.gas_particles.add_particles(gas)
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        mass = instance.gas_particles[0].mass
        instance.evolve_model(0.001 | generic_unit_system.time)
        self.assertEqual(instance.get_name_of_current_state(), 'EVOLVED')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()

        print("commit_parameters(), (re)commit_particles(), and cleanup_code() should be called "
            "automatically before new_xx_particle(), get_xx(), and stop():")
        instance = Gadget2(self.default_converter, **default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.gas_particles.add_particles(gas)
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.gas_particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.dm_particles.add_particles(dark)
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.gas_particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.evolve_model(0.001 | generic_unit_system.time)
        self.assertEqual(instance.get_name_of_current_state(), 'EVOLVED')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')

    def test11(self):
        particles = Particles(2)

        particles.x = [0.0, 10.0] | generic_unit_system.length
        particles.y = 0 | generic_unit_system.length
        particles.z = 0 | generic_unit_system.length
        particles.radius = 0.005 | generic_unit_system.length
        particles.vx = 0 | generic_unit_system.speed
        particles.vy = 0 | generic_unit_system.speed
        particles.vz = 0 | generic_unit_system.speed
        particles.mass = 1.0 | generic_unit_system.mass

        instance = Gadget2(self.default_converter, **few_particles_default_options)
        instance.initialize_code()
        instance.parameters.stopping_conditions_number_of_steps = 2
        self.assertEqual(instance.parameters.stopping_conditions_number_of_steps, 2)
        instance.parameters.epsilon_squared = (0.01 | generic_unit_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(1.0e15 | units.s)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        self.assertTrue(instance.model_time < 1.0e15 | units.s)

        instance.stop()

    def test12(self):
        print("Testing Gadget get_hydro_state_at_point")
        number_sph_particles = 100
        gas = new_evrard_gas_sphere(number_sph_particles, self.default_convert_nbody, seed=1234)
        instance = Gadget2(self.default_converter, **default_options)
        instance.parameters.n_smooth = 64
        instance.parameters.n_smooth_tol = 1.56250e-2
        instance.gas_particles.add_particles(gas)

        coords = [0.0 | units.kpc]*3
        hydro_state = instance.get_hydro_state_at_point(*coords)
        expected = [3.5469e-19 | units.kg * units.m**-3,
                            0.0 | units.kg * units.m**-2 / units.s,
                            0.0 | units.kg * units.m**-2 / units.s,
                            0.0 | units.kg * units.m**-2 / units.s,
                    (7.6297e-10 if testing_isotherm_no_gravity else
                     8.4252e-10) | units.kg * units.m**-1 * units.s**-2]
        for value, expect in zip(hydro_state, expected):
            self.assertAlmostRelativeEqual(value, expect, places=3)

        coords = [0.1 | units.kpc]*3
        hydro_state = instance.get_hydro_state_at_point(*coords)
        expected = [4.1456e-19 | units.kg * units.m**-3,
                            0.0 | units.kg * units.m**-2 / units.s,
                            0.0 | units.kg * units.m**-2 / units.s,
                            0.0 | units.kg * units.m**-2 / units.s,
                    (8.9175e-10 if testing_isotherm_no_gravity else
                     1.0742e-09) | units.kg * units.m**-1 * units.s**-2]
        for value, expect in zip(hydro_state, expected):
            self.assertAlmostRelativeEqual(value, expect, places=3)

        instance.stop()

    def test13(self):
        print("Testing Gadget get_hydro_state_at_point II: uniform sphere")
        number_sph_particles = 1000
        gas = new_uniform_spherical_particle_distribution(number_sph_particles, self.UnitLength, self.UnitMass)
        gas.velocity = [10.0, 20.0, 30.0] | units.km / units.s
        gas.u = 0.05 | generic_unit_system.specific_energy
        density = (1.0e10 | units.MSun) / (4.0/3.0 * numpy.pi * (1.0 | units.kpc)**3)

        instance = Gadget2(self.default_converter, **default_options)
        instance.parameters.n_smooth = 64
        instance.parameters.n_smooth_tol = 0.01
        instance.gas_particles.add_particles(gas)

        coords = [0.0 | units.kpc]*3
        rho, rhovx, rhovy, rhovz, rhoe = instance.get_hydro_state_at_point(*coords)
        self.assertAlmostRelativeEqual(rho,   density,                              places=3)
        self.assertAlmostRelativeEqual(rho,   max(instance.gas_particles.rho),      places=2)
        self.assertIsOfOrder(rho,   instance.gas_particles.rho)
        self.assertAlmostRelativeEqual(rhovx, density*instance.gas_particles[0].vx, places=3)
        self.assertAlmostRelativeEqual(rhovy, density*instance.gas_particles[0].vy, places=3)
        self.assertAlmostRelativeEqual(rhovz, density*instance.gas_particles[0].vz, places=3)
        self.assertAlmostRelativeEqual(rhoe,  density * (instance.gas_particles[0].u +
            0.5 * instance.gas_particles[0].velocity.length_squared()),  places=3)

        coords = [0.1 | units.kpc]*3
        rho, rhovx, rhovy, rhovz, rhoe = instance.get_hydro_state_at_point(*coords)
        self.assertAlmostRelativeEqual(rho,   density,                              places=3)
        self.assertAlmostRelativeEqual(rhovx, density*instance.gas_particles[0].vx, places=3)
        self.assertAlmostRelativeEqual(rhovy, density*instance.gas_particles[0].vy, places=3)
        self.assertAlmostRelativeEqual(rhovz, density*instance.gas_particles[0].vz, places=3)
        self.assertAlmostRelativeEqual(rhoe,  density * (instance.gas_particles[0].u +
            0.5 * instance.gas_particles[0].velocity.length_squared()),  places=3)
        instance.stop()

    def test14(self):
        print("Testing Gadget SPH particle properties")
        number_sph_particles = 1000
        gas = new_evrard_gas_sphere(number_sph_particles, self.default_convert_nbody, seed=1234)
        instance = Gadget2(self.default_converter, **default_options)
        instance.gas_particles.add_particles(gas)
        self.assertEqual(instance.gas_particles.num_neighbours >=
            instance.parameters.n_smooth*(1-instance.parameters.n_smooth_tol), True)
        self.assertEqual(instance.gas_particles.num_neighbours <=
            instance.parameters.n_smooth*(1+instance.parameters.n_smooth_tol), True)
        self.assertIsOfOrder(instance.gas_particles.h_smooth,
            self.default_convert_nbody.to_si(1.0 | nbody_system.length) *
            (instance.parameters.n_smooth*1.0/number_sph_particles)**(1.0/3))
        self.assertAlmostRelativeEqual(instance.gas_particles.u,
            self.default_convert_nbody.to_si(0.05 | nbody_system.specific_energy))

        # the density of the cloud scales with 1/r:
        r_sort, rho_sort = instance.gas_particles.position.lengths().sorted_with(instance.gas_particles.rho)
        mean_density = self.default_convert_nbody.to_si(3.0/(4.0*numpy.pi) | nbody_system.density)
        select = slice(number_sph_particles//2)  # select 50% particles closest to center to avoid boundaries
        self.assertIsOfOrder(rho_sort[select]/mean_density, r_sort.mean()/r_sort[select])
        self.assertAlmostEqual(instance.gas_particles.u * instance.gas_particles.rho,
            1.5 * instance.gas_particles.pressure)

        self.assertAlmostEqual(instance.gas_particles.du_dt, 0 | units.m**2 * units.s**-3)
        u_0 = instance.gas_particles.u.sum()
        instance.evolve_model(0.1 | units.Myr)
        if testing_isotherm_no_gravity:
            self.assertAlmostEqual(instance.gas_particles.du_dt, 0 | units.m**2 * units.s**-3)
        else:  # Collapsing ==> heating:
            self.assertTrue(instance.gas_particles.du_dt.sum() >= 0 | units.m**2 * units.s**-3)
            self.assertIsOfOrder(instance.gas_particles.du_dt.sum() * (0.1 | units.Myr),
                instance.gas_particles.u.sum() - u_0)
        instance.stop()

    def test15a(self):
        print("Test Gadget2 in periodic nogravity mode")
        instance = Gadget2(mode=Gadget2Interface.MODE_PERIODIC_NOGRAVITY,
            **few_particles_default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        # 'False' is default, but getting parameters implicitly calls initialize_code()...
        self.assertEqual(instance.parameters.periodic_boundaries_flag, True)
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        self.assertEqual(instance.parameters.no_gravity_flag, True)
        instance.parameters.periodic_box_size = 2.0 | generic_unit_system.length
        self.assertAlmostEqual(instance.parameters.periodic_box_size, 2.0 | units.kpc, places=6)
        instance.parameters.min_size_timestep = 1.0 | generic_unit_system.time

        instance.dm_particles.add_particles(self.three_particles_IC)
        self.assertAlmostEqual(instance.dm_particles.x, [0.5, 0., 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0., -0.5, 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0., 0., 0.5] | units.kpc, places=6)

        instance.evolve_model(0.1 | generic_unit_system.time)
        self.assertAlmostEqual(instance.dm_particles.x, [0.4, 0., 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0., -0.4, 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0., 0., 0.4] | units.kpc, places=6)

        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertAlmostEqual(instance.dm_particles.x, [-0.5, 0., 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0., 0.5, 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0., 0., -0.5] | units.kpc, places=6)
        instance.stop()

    def test15b(self):
        print("Test Gadget2 in nogravity mode")
        instance = Gadget2(mode="nogravity", **few_particles_default_options)
        self.assertEqual(instance.parameters.no_gravity_flag, True)
        self.assertEqual(instance.parameters.periodic_boundaries_flag, False)
        instance.parameters.min_size_timestep = 1.0 | generic_unit_system.time

        instance.dm_particles.add_particles(self.three_particles_IC)
        self.assertAlmostEqual(instance.dm_particles.x, [0.5, 0.0, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0.0, -0.5, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0.0, 0.0, 0.5] | units.kpc, places=6)

        instance.evolve_model(0.1 | generic_unit_system.time)
        self.assertAlmostEqual(instance.dm_particles.x, [0.4, 0.0, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0.0, -0.4, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0.0, 0.0, 0.4] | units.kpc, places=6)

        instance.evolve_model(1.0 | generic_unit_system.time)
        self.assertAlmostEqual(instance.dm_particles.x, [-0.5, 0.0, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0.0, 0.5, 0.0] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0.0, 0.0, -0.5] | units.kpc, places=6)
        instance.stop()

    def test15c(self):
        print("Test Gadget2 in periodic mode")
        converter = generic_unit_converter.ConvertBetweenGenericAndSiUnits(1.0 | units.yr, 1.0 | units.MSun, 1.0 | units.AU)
        instance = Gadget2(converter, mode=Gadget2Interface.MODE_PERIODIC_BOUNDARIES,
            **few_particles_default_options)
        self.assertEqual(instance.parameters.periodic_boundaries_flag, True)
        self.assertEqual(instance.parameters.no_gravity_flag, False)
        instance.parameters.periodic_box_size = 20.0 | generic_unit_system.length
        self.assertAlmostEqual(instance.parameters.periodic_box_size, 20.0 | units.AU, places=6)
        instance.parameters.max_size_timestep = 0.01 | units.yr
        instance.parameters.epsilon_squared = 1.0e-16 | units.RSun**2

        binary = Particles(2)
        binary.mass = 0.5 | units.MSun
        binary[1].position = [1.0, 0, 0] | units.AU
        binary.velocity = [0.0, 0, 0] | units.km / units.s
        binary[1].vy = (constants.G * binary.total_mass() / binary[1].x).sqrt()
        binary.move_to_center()

        binary.position += [7, -8, 3] | units.AU
        binary.velocity += [10, -10, 10] | units.AU / units.yr

        primary, secondary = instance.dm_particles.add_particles(binary)
        self.assertAlmostEqual(secondary.x - primary.x, 1.0 | units.AU, places=6)
        self.assertAlmostEqual(instance.dm_particles.center_of_mass(), [7, -8, 3] | units.AU, places=4)
        self.assertAlmostEqual(instance.dm_particles.center_of_mass_velocity(), [10, -10, 10] | units.AU / units.yr)

        instance.evolve_model(0.5 | units.yr)
        self.assertAlmostEqual(secondary.x - primary.x, -1.0 | units.AU, places=2)
        self.assertAlmostEqual(instance.dm_particles.center_of_mass(), [-8, 7, 8] | units.AU, places=4)

        instance.evolve_model(1.0 | units.yr)
        self.assertAlmostEqual(secondary.x - primary.x, 1.0 | units.AU, places=2)
        self.assertAlmostEqual(instance.dm_particles.center_of_mass(), [-3, 2, -7] | units.AU, places=4)
        instance.stop()

    def test16(self):
        instance = Gadget2(mode=Gadget2Interface.MODE_PERIODIC_BOUNDARIES,
            **few_particles_default_options)
        self.assertEqual(instance.parameters.periodic_boundaries_flag, True)
        instance.commit_parameters()  # its ok because read-only
        instance.dm_particles.add_particles(self.three_particles_IC)
        instance.stop()

    def test17(self):
        print("Testing Gadget parameters further")
        target_number_of_particles = 100
        gas = new_evrard_gas_sphere(target_number_of_particles, self.default_convert_nbody, seed=1234)
        instance = Gadget2(self.default_converter, **default_options)
        instance.parameters.time_limit_cpu = 0.001 | units.s
        instance.gas_particles.add_particles(gas)
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            1.0 | generic_unit_system.time,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -5, error is 'CPU-time limit reached.'"
            )
        )
        instance.parameters.time_limit_cpu = 10.0 | units.s
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            0.5 * instance.model_time,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -6, error is 'Can't evolve backwards in time.'"
            )
        )
        instance.stop()

    def test18(self):
        print("Testing Gadget parameters further")
        instance = Gadget2(mode=Gadget2Interface.MODE_PERIODIC_NOGRAVITY,
            **few_particles_default_options)
        instance.parameters.periodic_box_size = 2.0 | generic_unit_system.length  # implicitly calls initialize_code()...
        instance.parameters.begin_time = 10.04 | generic_unit_system.time
        instance.parameters.time_max = 10.4 | generic_unit_system.time
        instance.parameters.min_size_timestep = 0.1 | generic_unit_system.time

        self.assertAlmostEqual(instance.model_time, 0.0 | units.s)
        instance.dm_particles.add_particles(self.three_particles_IC)
        self.assertAlmostEqual(instance.model_time, instance.parameters.begin_time)
        self.assertAlmostEqual(instance.dm_particles.x, [0.5, 0., 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0., -0.5, 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0., 0., 0.5] | units.kpc, places=6)

        instance.evolve_model(10.14 | generic_unit_system.time)
        self.assertAlmostRelativeEqual(instance.model_time,
            instance.parameters.begin_time + instance.parameters.min_size_timestep, 7)
        self.assertAlmostEqual(instance.dm_particles.x, [0.4, 0., 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.y, [0., -0.4, 0.] | units.kpc, places=6)
        self.assertAlmostEqual(instance.dm_particles.z, [0., 0., 0.4] | units.kpc, places=6)
        instance.stop()

    def test19(self):
        particles = new_plummer_model(31)

        instance = Gadget2(self.default_converter, number_of_workers=1)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.01 | generic_unit_system.length ** 2
        instance.particles.add_particles(particles)
        instance.commit_particles()

        instance.evolve_model(0.001 | generic_unit_system.time)
        expected_positions = instance.particles.position
        instance.stop()
        positions_per_workers = []
        for n in [2, 3, 4, 5]:
            instance = Gadget2(self.default_converter, number_of_workers=n)
            instance.initialize_code()
            instance.parameters.epsilon_squared = 0.01 | generic_unit_system.length ** 2
            instance.commit_parameters()
            instance.particles.add_particles(particles)
            instance.commit_particles()

            instance.evolve_model(0.001 | generic_unit_system.time)
            positions_per_workers.append(instance.particles.position)
            instance.stop()

        for index, n in enumerate([2, 3, 4, 5]):
            self.assertAlmostRelativeEqual(expected_positions[:, 0], positions_per_workers[index][:, 0])
            self.assertAlmostRelativeEqual(expected_positions[:, 1], positions_per_workers[index][:, 1])
            self.assertAlmostRelativeEqual(expected_positions[:, 2], positions_per_workers[index][:, 2])

    def test20(self):
        print("Testing zero timestep exceptions")
        number_sph_particles = 1000
        gas = new_evrard_gas_sphere(number_sph_particles, self.default_convert_nbody, seed=1234)

        wrong_converter = generic_unit_converter.ConvertBetweenGenericAndSiUnits(
            self.UnitLength,
            self.UnitMass,
            1.0e15 * units.yr
        )
        instance = Gadget2(wrong_converter, **default_options)
        instance.gas_particles.add_particles(gas)
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            1.0e13 | units.yr,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -8, error is 'A particle was assigned a timestep"
                " of size zero. The code_time_unit used may be too large.'"
            )
        )
        instance.stop()

        instance = Gadget2(wrong_converter, **default_options)
        correct_converter = nbody_system.nbody_to_si(self.UnitLength, 1.0e15 * units.yr)
        gas_compatible_with_wrong_converter = new_evrard_gas_sphere(
            number_sph_particles,
            correct_converter,
            seed=1234
        )
        instance.gas_particles.add_particles(gas_compatible_with_wrong_converter)
        instance.evolve_model(1.0e13 | units.yr)
        instance.stop()

    def test21(self):
        print("Testing other evolve_model exceptions")
        number_sph_particles = 1000
        gas = new_evrard_gas_sphere(number_sph_particles, self.default_convert_nbody, seed=1234)

        wrong_converter = generic_unit_converter.ConvertBetweenGenericAndSiUnits(self.UnitLength,
            self.UnitMass, 1.0e-15 * units.yr)
        instance = Gadget2(wrong_converter, **default_options)
        instance.parameters.max_size_timestep = 3.0e-14 | units.yr
        instance.gas_particles.add_particles(gas)
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            1.1e-13 | units.yr,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -7, error is "
                "'Can't evolve further than time_max.'"
            )
        )
        instance.evolve_model(3.0e-14 | units.yr)
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            1.0e-14 | units.yr,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -6, error is 'Can't evolve backwards in time.'"
            )
        )
        instance.stop()

    def test22(self):
        print("Testing Gadget timestep limiter (Durier & Dalla Vecchia 2011)")
        number_sph_particles = 1000
        numpy.random.seed(345672)
        gas = new_plummer_gas_model(number_sph_particles, self.default_convert_nbody, base_grid=body_centered_grid_unit_cube)

        n_timescales = 0.025
        t_end = (n_timescales * self.default_convert_nbody.to_si(1.0 | nbody_system.time)).as_quantity_in(units.Myr)
        print("Evolving to ("+str(n_timescales), "nbody timescales): ", t_end)
        n_steps = 5
        times = (t_end * (range(1, n_steps+1)) / (1.0 * n_steps)).as_quantity_in(units.Myr)

        instance = Gadget2(self.default_converter, **default_options)
        instance.parameters.max_size_timestep = t_end
        gas = instance.gas_particles.add_particles(gas)

        kinetic_energies = [] | units.erg
        potential_energies = [] | units.erg
        thermal_energies = [] | units.erg

        for time in times[:3]:
            instance.evolve_model(time)
            potential_energies.append(instance.potential_energy)
            kinetic_energies.append(instance.kinetic_energy)
            thermal_energies.append(instance.thermal_energy)
            print(str(time)+": E =", potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1])

        central_particle = gas.select_array(lambda x, y, z: (x**2 + y**2 + z**2) == min(x**2 + y**2 + z**2), ["x", "y", "z"])
        print("Initial internal energy central particle:", central_particle[0].u.as_quantity_in(units.erg / units.g))
        delta_E = abs(instance.potential_energy) - central_particle[0].u * central_particle[0].mass
        central_particle.u = abs(instance.potential_energy) / central_particle.mass
        print("New internal energy central particle:", central_particle[0].u.as_quantity_in(units.erg / units.g))
        print("Changed total energy to:", potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1] + delta_E)

        self.assertAlmostRelativeEqual(
            instance.potential_energy + instance.kinetic_energy + instance.thermal_energy,
            potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1] + delta_E, 7)
        potential_energies[-1] = instance.potential_energy
        kinetic_energies[-1] = instance.kinetic_energy
        thermal_energies[-1] = instance.thermal_energy

        for i_step, time in enumerate(times[3:]):
            instance.evolve_model(time)
            potential_energies.append(instance.potential_energy)
            kinetic_energies.append(instance.kinetic_energy)
            thermal_energies.append(instance.thermal_energy)
            print(str(time)+": E =", potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1])
        instance.stop()

        total_energies = potential_energies + kinetic_energies + thermal_energies
        self.assertAlmostRelativeEqual(total_energies[1], total_energies[0], 3)
        self.assertAlmostRelativeEqual(total_energies[2], total_energies[1] + delta_E, 3)
        self.assertIsOfOrder(total_energies[3], total_energies[2])
        self.assertIsOfOrder(total_energies[4], total_energies[3])

    def slowtest22b(self):
        print("Testing Gadget timestep limiter (Durier & Dalla Vecchia 2011)")
        number_sph_particles = 10000
        numpy.random.seed(345672)
        gas = new_plummer_gas_model(number_sph_particles, self.default_convert_nbody, base_grid=body_centered_grid_unit_cube)

        n_pixels = 100
        n_timescales = 0.5
        t_end = (n_timescales * self.default_convert_nbody.to_si(1.0 | nbody_system.time)).as_quantity_in(units.Myr)
        print("Evolving to ("+str(n_timescales), "nbody timescales): ", t_end)
        n_steps = 100
        times = (t_end * (range(1, n_steps+1)) / (1.0 * n_steps)).as_quantity_in(units.Myr)

        instance = Gadget2(self.default_converter, **default_options)
        instance.parameters.max_size_timestep = t_end
        gas = instance.gas_particles.add_particles(gas)

        kinetic_energies = [] | units.erg
        potential_energies = [] | units.erg
        thermal_energies = [] | units.erg

        for time in times[:3]:
            instance.evolve_model(time)
            potential_energies.append(instance.potential_energy)
            kinetic_energies.append(instance.kinetic_energy)
            thermal_energies.append(instance.thermal_energy)
            print(time, potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1])

        view = ([-0.5, 0.5, -0.5, 0.5] * (2.0 | units.none) *
            self.default_convert_nbody.to_si(1.0 | nbody_system.length)).as_quantity_in(units.kpc)
        print("Physical size of window for hydro_plot:", view)
        hydro_plot(
            view,
            instance,
            (n_pixels, n_pixels),
            "gadget2_slowtest22b_hydro_image{0:=03}.png".format(0)
        )
        central_particle = gas.select_array(lambda x, y, z: (x**2 + y**2 + z**2) == min(x**2 + y**2 + z**2), ["x", "y", "z"])
        print("Position central particle:", central_particle.position.as_quantity_in(units.kpc))
        print("Center of mass, virial radius of system:", gas.center_of_mass().as_quantity_in(units.kpc), gas.virial_radius().as_quantity_in(units.kpc))
        print("Initial internal energy central particle:", central_particle[0].u.as_quantity_in(units.erg / units.g))
        delta_E = abs(instance.potential_energy) - central_particle[0].u * central_particle[0].mass
        central_particle.u = abs(instance.potential_energy) / central_particle.mass
        print("New internal energy central particle:", central_particle[0].u.as_quantity_in(units.erg / units.g))
        print("Changed total energy to:", potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1] + delta_E)

        self.assertAlmostRelativeEqual(
            instance.potential_energy + instance.kinetic_energy + instance.thermal_energy,
            potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1] + delta_E, 7)
        potential_energies[-1] = instance.potential_energy
        kinetic_energies[-1] = instance.kinetic_energy
        thermal_energies[-1] = instance.thermal_energy
        print(potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1])

        for i_step, time in enumerate(times[3:]):
            if i_step < 20:
                hydro_plot(
                    view,
                    instance,
                    (n_pixels, n_pixels),
                    "gadget2_slowtest22b_hydro_image{0:=03}.png".format(i_step+1)
                )
            instance.evolve_model(time)
            potential_energies.append(instance.potential_energy)
            kinetic_energies.append(instance.kinetic_energy)
            thermal_energies.append(instance.thermal_energy)
            print(time, potential_energies[-1] + kinetic_energies[-1] + thermal_energies[-1])
        instance.stop()

        total_energies = potential_energies + kinetic_energies + thermal_energies
        self.assertAlmostRelativeEqual(total_energies[1], total_energies[0], 3)
        self.assertAlmostRelativeEqual(total_energies[2], total_energies[1] + delta_E, 3)
        self.assertIsOfOrder(total_energies[3], total_energies[2])
        self.assertIsOfOrder(total_energies[4], total_energies[3])
        for i in range(5, n_steps):
            self.assertAlmostRelativeEqual(total_energies[i], total_energies[i-1], 1)

        energy_evolution_plot(times, kinetic_energies, potential_energies, thermal_energies,
            figname="gadget2_slowtest22b_energy.png")

    def test23(self):
        print("Testing Gadget2 density_limit_detection")
        number_gas_particles = 1000
        gas = new_evrard_gas_sphere(number_gas_particles, self.default_convert_nbody, do_scale=True, seed=12345)

        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        instance.parameters.stopping_condition_maximum_density = 10 * self.UnitMass / self.UnitLength**3
        instance.gas_particles.add_particles(gas)
        stars = new_plummer_model(5, self.default_convert_nbody)
        stars.x += 1000 * self.UnitLength
        instance.dm_particles.add_particles(stars)
        self.assertIsOfOrder(max(instance.gas_particles.density), self.UnitMass / self.UnitLength**3)

        density_limit_detection = instance.stopping_conditions.density_limit_detection
        density_limit_detection.enable()

        instance.evolve_model(10.0 | units.Myr)
        print(instance.model_time.as_quantity_in(units.Myr))
        print(instance.stopping_conditions)
        self.assertTrue(density_limit_detection.is_set())
        self.assertTrue(instance.model_time < 10.0 | units.Myr)
        self.assertEqual(len(density_limit_detection.particles()), 1)
        self.assertTrue(density_limit_detection.particles().density >
                10 * self.UnitMass / self.UnitLength**3)

        instance.particles.remove_particles(density_limit_detection.particles())

        instance.evolve_model(10.0 | units.Myr)
        print(instance.model_time.as_quantity_in(units.Myr))
        print(instance.stopping_conditions)
        self.assertTrue(density_limit_detection.is_set())
        self.assertTrue(instance.model_time < 10.0 | units.Myr)

        self.assertEqual(len(density_limit_detection.particles()), 2)
        self.assertEqual((density_limit_detection.particles().density >
                10 * self.UnitMass / self.UnitLength**3), [True, True])
        instance.stop()

    def test24(self):
        print("Testing Gadget2 internal_energy_limit_detection")
        number_gas_particles = 1000
        gas = new_evrard_gas_sphere(number_gas_particles, self.default_convert_nbody, do_scale=True, seed=12345)
        initial_internal_energy = 0.05 * constants.G * self.UnitMass / self.UnitLength

        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        instance.parameters.stopping_condition_maximum_internal_energy = 10 * initial_internal_energy
        instance.gas_particles.add_particles(gas)
        self.assertAlmostRelativeEquals(instance.gas_particles.u, initial_internal_energy, 8)

        internal_energy_limit_detection = instance.stopping_conditions.internal_energy_limit_detection
        internal_energy_limit_detection.enable()

        instance.evolve_model(10.0 | units.Myr)
        print(instance.model_time.as_quantity_in(units.Myr))
        print(instance.stopping_conditions)
        self.assertTrue(internal_energy_limit_detection.is_set())
        self.assertTrue(instance.model_time < 10.0 | units.Myr)
        self.assertEqual(len(internal_energy_limit_detection.particles()), 4)
        self.assertEqual((internal_energy_limit_detection.particles().u >
                10 * initial_internal_energy), [True, True, True, True])

        instance.particles.remove_particles(internal_energy_limit_detection.particles())

        instance.evolve_model(10.0 | units.Myr)
        print(instance.model_time.as_quantity_in(units.Myr))
        print(instance.stopping_conditions)
        self.assertTrue(internal_energy_limit_detection.is_set())
        self.assertTrue(instance.model_time < 10.0 | units.Myr)
        self.assertEqual(len(internal_energy_limit_detection.particles()), 3)
        self.assertEqual((internal_energy_limit_detection.particles().u >
                10 * initial_internal_energy), [True, True, True])
        instance.stop()

    def test25(self):
        print("Testing Gadget2 comoving_integration_flag")
        particles = new_plummer_model(100, self.default_convert_nbody, do_scale=True)
        instance = Gadget2(self.default_converter, **default_options)
        instance.initialize_code()
        instance.parameters.comoving_integration_flag = True
        instance.parameters.redshift_begin = 20.0
        instance.parameters.redshift_max = 0.0
        instance.parameters.omega_zero = 0.3
        instance.parameters.omega_lambda = 0.7
        instance.parameters.omega_baryon = 0.0
        instance.parameters.hubble_parameter = 0.7
        instance.parameters.epsilon_squared = (1.0 | generic_unit_system.length)**2
        instance.parameters.softening_halo_max_phys = 1.0 | generic_unit_system.length
        instance.dm_particles.add_particles(particles)

        instance.evolve_to_redshift(19.9)
        self.assertRaises(
            AmuseException,
            instance.evolve_model,
            1 | units.Myr,
            expected_message=(
                f"Error when calling 'evolve_model' of a '{self.classname}', "
                "errorcode is -9, error is 'This function should not be used "
                "with the current value of comoving_integration_flag'"
            )
        )

        self.assertAlmostEqual(instance.model_redshift, 19.9, 5)
        self.assertRaises(
            AmuseException,
            getattr,
            instance,
            "model_time",
            expected_message=(
                f"Error when calling 'get_time' of a '{self.classname}', "
                "errorcode is -9, error is 'This function should not be used "
                "with the current value of comoving_integration_flag'"
            )
        )
        instance.stop()

    def xtest26(self):
        print("Testing Gadget2 comoving_integration_flag validation (WIP)")
        particles = read_set_from_file("cluster_littleendian.dat", "gadget")
        particles = ParticlesSuperset(particles[1:4])
        instance = Gadget2(self.default_converter, redirection="none", **default_options)
        instance.initialize_code()
        instance.parameters.comoving_integration_flag = True
        instance.parameters.redshift_begin = 23.0
        instance.parameters.redshift_max = -0.1
        instance.parameters.omega_zero = 0.3
        instance.parameters.omega_lambda = 0.7
        instance.parameters.omega_baryon = 0.0
        instance.parameters.hubble_parameter = 0.7
        instance.parameters.epsilon_squared = (72.0 | generic_unit_system.length)**2
        instance.parameters.softening_halo_max_phys = 12.0 | generic_unit_system.length
        instance.parameters.max_size_timestep = 0.03 | generic_unit_system.time
        instance.parameters.tree_domain_update_frequency = 0.1
        instance.particles.add_particles(particles)
        instance.evolve_to_redshift(5.0)
        write_set_to_file(instance.particles, "z5", "hdf5")
        instance.evolve_to_redshift(3.0)
        write_set_to_file(instance.particles, "z3", "hdf5")
        instance.evolve_to_redshift(2.0)
        write_set_to_file(instance.particles, "z2", "hdf5")
        instance.evolve_to_redshift(1.0)
        write_set_to_file(instance.particles, "z1", "hdf5")
        instance.evolve_to_redshift(0.0)
        write_set_to_file(instance.particles, "z0", "hdf5")
        instance.stop()

    def xtest27(self):
        print("Testing Gadget2 comoving_integration_flag validation II (WIP)")
        particles = read_set_from_file("lcdm_gas_littleendian.dat", "gadget")
        gas = particles[0]
        mu = constants.proton_mass * 4.0 / (1.0 + 3.0 * 0.76)
        gas.u = self.default_converter.to_generic(3 * constants.kB * (1000 | units.K) / (2 * mu))
        gas.velocity /= 11.0**-0.5  # velocities in Gadget files are in w=sqrt(a)*dx/dt
        halo = particles[1]
        halo.velocity /= 11.0**-0.5  # velocities in Gadget files are in w=sqrt(a)*dx/dt
        gas = convert_comoving_to_physical(gas, redshift=10.0)
        halo = convert_comoving_to_physical(halo, redshift=10.0)

        instance = Gadget2(self.default_converter, mode="periodic", redirection="none", number_of_workers=1)
        instance.parameters.comoving_integration_flag = True
        instance.parameters.n_smooth = 33.0
        instance.parameters.n_smooth_tol = 2.0 / instance.parameters.n_smooth
        instance.parameters.periodic_box_size = 50000.0 | generic_unit_system.length
        instance.parameters.redshift_begin = 10.0
        instance.parameters.redshift_max = -0.1
        instance.parameters.omega_zero = 0.3
        instance.parameters.omega_lambda = 0.7
        instance.parameters.omega_baryon = 0.04
        instance.parameters.hubble_parameter = 1.0  # coordinates in input file are already in kpc/h
        instance.parameters.artificial_viscosity_alpha = 0.8
        instance.parameters.min_gas_temp = 50.0 | units.K
        instance.parameters.gas_epsilon = 600.0 | generic_unit_system.length
        instance.parameters.epsilon_squared = (600.0 | generic_unit_system.length)**2
        instance.parameters.softening_gas_max_phys = 600.0 | generic_unit_system.length
        instance.parameters.softening_halo_max_phys = 600.0 | generic_unit_system.length
        instance.parameters.max_size_timestep = 0.03 | generic_unit_system.time
        instance.parameters.tree_domain_update_frequency = 0.1
        instance.gas_particles.add_particles(gas)
        instance.dm_particles.add_particles(halo)
        write_set_to_file(instance.particles, "z10", "hdf5")
        instance.evolve_to_redshift(5.0)
        write_set_to_file(instance.particles, "z5", "hdf5")
        instance.evolve_to_redshift(3.0)
        write_set_to_file(instance.particles, "z3", "hdf5")
        instance.evolve_to_redshift(2.0)
        write_set_to_file(instance.particles, "z2", "hdf5")
        instance.evolve_to_redshift(1.0)
        write_set_to_file(instance.particles, "z1", "hdf5")
        instance.evolve_to_redshift(0.0)
        write_set_to_file(instance.particles, "z0", "hdf5")
        instance.stop()

    def test28(self):
        print("Testing Gadget get_hydro_state_at_point in periodic mode")
        gas = Particles(8000)
        gas.mass = 1.0/8000 | generic_unit_system.mass
        x, y, z = numpy.mgrid[-1:0.9:20j, -1:0.9:20j, -1:0.9:20j] | units.kpc
        gas.x, gas.y, gas.z = x.flatten(), y.flatten(), z.flatten()
        gas.velocity = [0, 0, 0] | units.m / units.s
        gas.u = (1 + 0.8 * numpy.sin(gas.x * (numpy.pi | units.kpc**-1))) | generic_unit_system.specific_energy

        instance = Gadget2(self.default_converter, mode='periodic', **default_options)
        instance.parameters.periodic_box_size = 2.0 | units.kpc
        instance.gas_particles.add_particles(gas)

        self.assertAlmostRelativeEqual(instance.total_mass, 1.0e10 | units.MSun, 3)
        self.assertAlmostRelativeEqual(instance.thermal_energy, 1.0e10 | units.MSun * units.km**2 * units.s**-2, 3)

        number_of_points = 100
        in_domain = numpy.linspace(-1.0, 1.0, 100) | units.kpc
        domain_border = numpy.ones(100) | units.kpc
        state_left = instance.get_hydro_state_at_point(-domain_border, in_domain, in_domain)
        state_right = instance.get_hydro_state_at_point(domain_border, in_domain, in_domain)
        for var_left, var_right in zip(state_left, state_right):
            self.assertEqual(var_left, var_right)

        state_back = instance.get_hydro_state_at_point(in_domain, -domain_border, in_domain)
        state_front = instance.get_hydro_state_at_point(in_domain, domain_border, in_domain)
        for var_front, var_back in zip(state_front, state_back):
            self.assertEqual(var_front, var_back)

        state_bottom = instance.get_hydro_state_at_point(in_domain, in_domain, -domain_border)
        state_top = instance.get_hydro_state_at_point(in_domain, in_domain, domain_border)
        for var_top, var_bottom in zip(state_top, state_bottom):
            self.assertEqual(var_top, var_bottom)

        self.assertAlmostRelativeEqual(state_left[0].mean(), 1.25e9 | units.MSun * units.kpc**-3, 2)
        self.assertAlmostRelativeEqual(state_back[0].mean(), 1.25e9 | units.MSun * units.kpc**-3, 2)
        self.assertAlmostRelativeEqual(state_bottom[0].mean(), 1.25e9 | units.MSun * units.kpc**-3, 2)

        self.assertAlmostRelativeEqual(state_left[4].mean(), 1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2, 2)
        self.assertAlmostRelativeEqual(state_back[4].mean(), 1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2, 2)
        self.assertAlmostRelativeEqual(state_bottom[4].mean(), 1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2, 2)

        # With x=cst, variations in u (and therefore in rhoe) should be small
        self.assertAlmostEqual(state_left[4].std() / (1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2), 0.0, 2)
        # Over the entire x domain, variations in rhoe should equal the sine amplitude (=0.8) times the rms(sine) over a whole period (=sqrt(.5))
        self.assertAlmostEqual(state_back[4].std() / (1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2), 0.8 * numpy.sqrt(0.5), 1)
        self.assertAlmostEqual(state_bottom[4].std() / (1.25e9 | units.MSun * units.kpc**-3 * units.km**2 * units.s**-2), 0.8 * numpy.sqrt(0.5), 1)
        instance.stop()


def energy_evolution_plot(time, kinetic, potential, thermal, figname):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(5, 5))
    plot(time, kinetic, label='K')
    plot(time, potential, label='U')
    plot(time, thermal, label='Q')
    plot(time, kinetic + potential + thermal, label='E')
    xlabel('Time')
    ylabel('Energy')
    pyplot.legend(prop={'size': "x-small"}, loc=4)
    pyplot.savefig(figname)
    print("\nPlot of energy evolution was saved to:", figname)
    pyplot.close()


def hydro_plot(view, hydro_code, image_size, figname):
    """
    view: the (physical) region to plot [xmin, xmax, ymin, ymax]
    hydro_code: hydrodynamics code in which the gas to be plotted is defined
    image_size: size of the output image in pixels (x, y)
    """
    if not HAS_MATPLOTLIB:
        return
    shape = (image_size[0], image_size[1], 1)
    size = image_size[0] * image_size[1]
    axis_lengths = [0.0, 0.0, 0.0] | units.m
    axis_lengths[0] = view[1] - view[0]
    axis_lengths[1] = view[3] - view[2]
    grid = new_regular_grid(shape, axis_lengths)
    grid.x += view[0]
    grid.y += view[2]
    speed = grid.z.reshape(size) * (0 | 1/units.s)
    rho, rhovx, rhovy, rhovz, rhoe = hydro_code.get_hydro_state_at_point(grid.x.reshape(size),
        grid.y.reshape(size), grid.z.reshape(size))

    min_v = 800.0 | units.km / units.s
    max_v = 3000.0 | units.km / units.s
    min_rho = rho.amin()
    max_rho = rho.amax()
    fade_rho = min_rho * (max_rho / min_rho)**0.1
    min_E = 1.0e11 | units.J / units.kg
    max_E = 1.0e13 | units.J / units.kg

    v_sqr = (rhovx**2 + rhovy**2 + rhovz**2) / rho**2
    E = rhoe / rho
    log_v = numpy.log((v_sqr / min_v**2)) / numpy.log((max_v**2 / min_v**2))
    log_rho = numpy.log((rho / min_rho)) / numpy.log((max_rho / min_rho))
    log_E = numpy.log((E / min_E)) / numpy.log((max_E / min_E))

    red = numpy.minimum(numpy.ones_like(rho.number), numpy.maximum(numpy.zeros_like(rho.number), log_rho)).reshape(shape)
    green = numpy.minimum(numpy.ones_like(rho.number), numpy.maximum(numpy.zeros_like(rho.number), log_v)).reshape(shape)
    blue = numpy.minimum(numpy.ones_like(rho.number), numpy.maximum(numpy.zeros_like(rho.number), log_E)).reshape(shape)
    alpha = numpy.minimum(numpy.ones_like(log_v), numpy.maximum(numpy.zeros_like(log_v),
        numpy.log((rho / min_rho)) / numpy.log((fade_rho / min_rho)))).reshape(shape)

    rgba = numpy.concatenate((red, green, blue, alpha), axis=2)

    pyplot.figure(figsize=(image_size[0]/100.0, image_size[1]/100.0), dpi=100)
    im = pyplot.figimage(rgba, origin='lower')

    pyplot.savefig(figname, transparent=True, dpi=100)
    print("\nHydroplot was saved to: ", figname)
    pyplot.close()



from amuse.units import nbody_system
from amuse.units import units
from amuse.test.amusetest import get_path_to_results
from amuse.test import compile_tools
import subprocess
import os
import numpy
import time

from amuse import datamodel
from amuse.rfi.tools import create_c
from amuse.rfi import channel
from amuse.rfi.core import *
codestring = """

#include <stdio.h>
#include <mpi.h>
#include <new>
#include <iostream>


struct data {
    double x;
    double y;
    double z;
};

int number_of_points_in_one_dimension = 0;
data * model = 0;

int set_data(int index, double vx, double vy, double vz)
{
    if(!model)
    {
        return -1;
    }
    
    if(index > (number_of_points_in_one_dimension * number_of_points_in_one_dimension * number_of_points_in_one_dimension))
    {
        return -2;
    }
    data & m = model[index];
    m.x = vx;
    m.y = vy;
    m.z = vz;
    return 0;
}


int get_data(int index, double * vx, double  * vy, double  * vz)
{
    double data_in[6], data_out[6];
    int status_in,status_out;
    if(!model)
    {
        return -1;
    }
    
    if(index > (number_of_points_in_one_dimension * number_of_points_in_one_dimension * number_of_points_in_one_dimension))
    {
        return -2;
    }
    data & m = model[index];
    *vx = m.x;
    *vy = m.y;
    *vz = m.z;
    data_in[0] = data_in[1] = data_in[2] = 0.0;
    data_in[3] = data_in[4] = data_in[5] = 0.0;
    data_out[0] = data_out[1] = data_out[2] = 0.0;
    data_out[3] = data_out[4] = data_out[5] = 0.0;
    status_in = status_out = 0;
    /*
    MPI::COMM_WORLD.Allreduce(data_in, data_out, 6, MPI::DOUBLE,MPI::SUM);
    MPI::COMM_WORLD.Barrier();
    MPI::COMM_WORLD.Allreduce(&status_in, &status_out, 1, MPI::DOUBLE,MPI::SUM);
    */
    return 0;
}
  
int step() 
{
    if(!model) {
        return -1;
    }
    for(int xindex ; xindex < number_of_points_in_one_dimension; xindex++)
    {
        for(int yindex ; yindex < number_of_points_in_one_dimension; yindex++)
        {
            for(int zindex ; zindex < number_of_points_in_one_dimension; zindex++)
            {
                int index = xindex * number_of_points_in_one_dimension * number_of_points_in_one_dimension;
                index += yindex * number_of_points_in_one_dimension;
                index += zindex;
                
                model[index].x = index;
                model[index].y = model[index].x / (1.0 + model[index].y);
                model[index].z = model[index].x  * model[index].y / (model[index].z + 1e-7);
            }
        }
    }
    return 0;
}    
  
int set_number_of_points_in_one_dimension(int value)
{
    if(model) {
        delete model;
    }
    
    try {
        model = new data[value*value*value];
    } catch (std::bad_alloc &e) {
        number_of_points_in_one_dimension = 0;
        return -1;
    }
    number_of_points_in_one_dimension = value;
    
    return 0;
    
}

int set_data_to_same(int n, double vx, double vy, double vz) {
    for(int i = 0; i < n; i++) {
        set_data(i, vx, vy, vz);
    }
    return 0;
}

int reset()
{
    if(model) {
        delete model;
    }
    model = 0;
    return 0;
}
"""

class TestCode(CodeInterface):
    
    def __init__(self, exefile):
        CodeInterface.__init__(self, exefile)
         
         
    @legacy_function
    def set_number_of_points_in_one_dimension():
        """
        Set the set number of points in one dimension (N), the total model
        size will be qubed (N*N*N)
        """
        function = LegacyFunctionSpecification()  
        function.addParameter('value',
            dtype='int32',
            direction=function.IN,
            description =  
                "The number of points in one direction")
        function.result_type = 'int32'
        return function  
        
    @legacy_function
    def step():
        """
        Do one step over the N * N * N grid
        """
        function = LegacyFunctionSpecification()  
        function.result_type = 'int32'
        return function
        
    
    @legacy_function
    def reset():
        """
        Restore the model to its original state
        """
        function = LegacyFunctionSpecification()  
        function.result_type = 'int32'
        return function  
        
    @legacy_function
    def set_data():
        """
        set example vector data
        """
        function = LegacyFunctionSpecification()  
        function.addParameter('index',
            dtype='int32',
            direction=function.IN,
            description =  
                "index in the array in range 0 <= index < (N*3)")
        function.addParameter('vx',
            dtype='float64',
            direction=function.IN,
            description =  
                "x component of the vector")
        function.addParameter('vy',
            dtype='float64',
            direction=function.IN,
            description =  
                "y component of the vector")
        function.addParameter('vz',
            dtype='float64',
            direction=function.IN,
            description =  
                "z component of the vector")
        function.can_handle_array = True
        function.result_type = 'int32'
        return function 
        
    @legacy_function
    def set_data_to_same():
        """
        set all vector data to same value
        """
        function = LegacyFunctionSpecification()  
        function.addParameter('max',
            dtype='int32',
            direction=function.IN,
            description =  
                "index in the array in range 0 <= index < (N*3)")
        function.addParameter('vx',
            dtype='float64',
            direction=function.IN,
            description =  
                "x component of the vector")
        function.addParameter('vy',
            dtype='float64',
            direction=function.IN,
            description =  
                "y component of the vector")
        function.addParameter('vz',
            dtype='float64',
            direction=function.IN,
            description =  
                "z component of the vector")
        function.can_handle_array = True
        function.result_type = 'int32'
        return function
        
    @legacy_function
    def get_data():
        """
        retrieve example vector data
        """
        function = LegacyFunctionSpecification()  
        function.addParameter('index',
            dtype='int32',
            direction=function.IN,
            description =  
                "index in the array in range 0 <= index < (N*3)")
        function.addParameter('vx',
            dtype='float64',
            direction=function.OUT,
            description =  
                "x component of the vector")
        function.addParameter('vy',
            dtype='float64',
            direction=function.OUT,
            description =  
                "y component of the vector")
        function.addParameter('vz',
            dtype='float64',
            direction=function.OUT,
            description =  
                "z component of the vector")
        function.can_handle_array = True
        function.result_type = 'int32'
        return function  



class RunSpeedTests(object):
    
    def __init__(self):
        self.number_of_gridpoints = [8]            
    
    def build_worker(self):
        
        path = os.path.abspath(get_path_to_results())
        codefile = os.path.join(path,"code.o")
        interfacefile = os.path.join(path,"interface.o")
        self.exefile = os.path.join(path,"c_worker")
        
        compile_tools.cxx_compile(codefile, codestring)
        
        uc = create_c.GenerateACHeaderStringFromASpecificationClass()
        uc.specification_class = TestCode
        uc.make_extern_c = False
        header =  uc.result
        
        
        uc = create_c.GenerateACSourcecodeStringFromASpecificationClass()
        uc.specification_class = TestCode
        uc.needs_mpi=False
        code =  uc.result
        
        string = '\n\n'.join([header, code])
        
        #print string
        
        compile_tools.cxx_compile(interfacefile, string, extra_args=['-I' , path])
        compile_tools.c_build(self.exefile, [interfacefile, codefile] )
    
    def start(self):
        self.build_worker()
        
        
        for number_of_points_in_one_dimension in self.number_of_gridpoints:
            result = self.run(number_of_points_in_one_dimension)
    
            print(', '.join([str(x) for x in result]))
                
    def run(self, number_of_points_in_one_dimension):
    
        instance = TestCode(self.exefile)

        total_number_of_points = number_of_points_in_one_dimension ** 3
        number_of_bytes = 4 + 8 + 8 + 8
        total_number_of_bytes = total_number_of_points * (number_of_bytes + 4)
        indices = numpy.array(range(total_number_of_points), dtype='int32')

        data_x = numpy.array(range(total_number_of_points), dtype='float64')
        data_y = numpy.array(range(total_number_of_points), dtype='float64')
        data_z = numpy.array(range(total_number_of_points), dtype='float64')

        errorcode = instance.set_number_of_points_in_one_dimension(number_of_points_in_one_dimension)
        if errorcode < 0:
            raise Exception("Could not allocate memory")

        t0 = time.time()
        instance.set_data(indices, data_x, data_y, data_z)
        t1 = time.time()
        dt = t1 - t0
        mbytes_per_second = total_number_of_bytes / dt / (1000.0 * 1000.0)

        t2 = time.time()
        instance.set_data_to_same(total_number_of_points, 0.0, 1.0, 2.0)
        t3 = time.time()
        
        instance.reset()
        instance.stop()
        
        return dt, total_number_of_points, mbytes_per_second, t3-t2, (dt - (t3-t2)) / (t3-t2)     
        
        
def test_speed():
    x = RunSpeedTests()
    x.number_of_gridpoints = [8]
    x.start()

if __name__ == '__main__':
    #channel.MessageChannel.DEBUGGER = channel.MessageChannel.DDD
    x = RunSpeedTests()
    x.number_of_gridpoints = [64, 128, 192]
    x.start()



from amuse.test import amusetest
from amuse.support.exceptions import AmuseException

import os
import tempfile
from amuse import io
from amuse.io import base
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel


class ForTestFileFormatProcessor(base.FileFormatProcessor):
    """
    Save files in a test format

    long description    
    """
    provided_formats = ['test', '123']
    instance = None

    def __init__(self, filename=None, set=None, format=None):
        base.FileFormatProcessor.__init__(self, filename, set, format)
        ForTestFileFormatProcessor.instance = self
        self.stored = False

    @base.format_option
    def add_comma(self):
        """if True will add a comma between each value"""
        return True

    @base.format_option
    def save_fast(self):
        """if True will save faster but less accurate"""
        return False

    def store(self):
        self.stored = True

    def load(self):
        return 10


class FrameworkTests(amusetest.TestCase):

    def test1(self):
        options = ForTestFileFormatProcessor.get_options()
        self.assertTrue('add_comma' in options)
        self.assertTrue('save_fast' in options)
        ForTestFileFormatProcessor.register()
        base.write_set_to_file(None, "test.txt", format="123")
        self.assertEqual(ForTestFileFormatProcessor.instance.set, None)
        self.assertEqual(ForTestFileFormatProcessor.instance.filename, "test.txt")
        self.assertEqual(ForTestFileFormatProcessor.instance.format, "123")
        self.assertTrue(ForTestFileFormatProcessor.instance.stored)
        self.assertTrue(ForTestFileFormatProcessor.instance.add_comma)

    def test2(self):
        ForTestFileFormatProcessor.register()
        base.write_set_to_file(None, "test.txt", format="123", add_comma=False)
        self.assertFalse(ForTestFileFormatProcessor.instance.add_comma)

    def test3(self):
        ForTestFileFormatProcessor.register()
        documentation = base.write_set_to_file.__doc__
        self.assertTrue("**123**,\n      Save files in a test format" in documentation)

    def test4(self):
        options = base.get_options_for_format('123')
        options.sort(key=lambda x: x[0])
        name, description, default = options[0]
        self.assertEqual(name, 'add_comma')
        self.assertEqual(description, 'if True will add a comma between each value')
        self.assertEqual(default, True)
        name, description, default = options[2]
        self.assertEqual(name, 'save_fast')
        self.assertEqual(description, 'if True will save faster but less accurate')
        self.assertEqual(default, False)

    def test5(self):
        self.assertRaises(AmuseException, io.read_set_from_file, "non_existent", "test",
            expected_message="IO exception: Error: file 'non_existent' does not exist.")

        processor = base.FileFormatProcessor(format="test")
        self.assertRaises(base.CannotSaveException, processor.store,
            expected_message="You tried to save a file with fileformat 'test', but"
                " this format is not supported for writing files")

        self.assertRaises(base.CannotLoadException, processor.load,
            expected_message="You tried to load a file with fileformat 'test', but"
                " this format is not supported for reading files")


class FormatTests(amusetest.TestCase):

    def test1(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | nbody_system.mass
        x.radius = [3.0, 4.0] | nbody_system.length
        x.position = [[1, 2, 3], [3, 5, 6]] | nbody_system.length
        x.velocity = [[1, 2, 3], [3, 5, 6]] | nbody_system.speed
        io.write_set_to_file(x, "test.tsf", "tsf")
        y = io.read_set_from_file("test.tsf", "tsf")

        self.assertAlmostEqual(x.mass, y.mass, 8)
#        self.assertAlmostEquals(x.radius, y.radius, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        os.remove("test.tsf")

    def test2(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | nbody_system.mass
        x.radius = [3.0, 4.0] | nbody_system.length
        x.position = [[1, 2, 3], [3, 5, 6]] | nbody_system.length
        x.velocity = [[1, 2, 3], [3, 5, 6]] | nbody_system.speed
        io.write_set_to_file(x, "test.dyn", "dyn")
        y = io.read_set_from_file("test.dyn", "dyn")

        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)
        self.assertRaises(AttributeError, lambda: y.radius,
            expected_message="You tried to access attribute 'radius' but this "
                "attribute is not defined for this set.")

        os.remove("test.dyn")

    def test3(self):
        x = datamodel.Particles(2)
        convert = nbody_system.nbody_to_si(1 | units.kg, 2 | units.m)
        x.mass = [1.0, 2.0] | units.kg
        x.radius = [3.0, 4.0] | units.m
        x.position = [[1, 2, 3], [3, 5, 6]] | units.m
        x.velocity = [[1, 2, 3], [3, 5, 6]] | units.m / units.s
        io.write_set_to_file(x, "test_unit.dyn", "dyn", nbody_to_si_converter=convert)
        y = io.read_set_from_file("test_unit.dyn", "dyn", nbody_to_si_converter=convert)

        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)
        self.assertRaises(AttributeError, lambda: y.radius,
            expected_message="You tried to access attribute 'radius' but this "
                "attribute is not defined for this set.")

        os.remove("test_unit.dyn")

    def test4(self):

        x = datamodel.Particles(2)
        convert = nbody_system.nbody_to_si(1 | units.kg, 2 | units.m)
        x.mass = [1.0, 2.0] | units.kg
        x.radius = [3.0, 4.0] | units.m
        x.position = [[1, 2, 3], [3, 5, 6]] | units.m
        x.velocity = [[1, 2, 3], [3, 5, 6]] | units.m / units.s
        io.write_set_to_file(x, "test_unit.tsf", "tsf", nbody_to_si_converter=convert)
        y = io.read_set_from_file("test_unit.tsf", "tsf", nbody_to_si_converter=convert)

        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)
        self.assertRaises(AttributeError, lambda: y.radius,
            expected_message="You tried to access attribute 'radius' but this "
                "attribute is not defined for this set.")

        os.remove("test_unit.tsf")

    def test5(self):
        print("Testing HDF5 io")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        x.radius = [3.0, 4.0] | units.m
        x.position = [[1, 2, 3], [3, 5, 6]] | units.m
        x.velocity = [[1, 2, 3], [3, 5, 6]] | units.m / units.s
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5")

        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.radius, y.radius, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        os.remove("test_unit.hdf5")

    def test6(self):
        print("Testing HDF5 io, with options")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="1.0")
        x.mass = [10.0, 20.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="1.0")
        x.mass = [100.0, 200.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="1.0", append_to_file=True)
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", copy_history=False, close_file=False)
        y = y.previous_state()  # weirdness for version="1.0"
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual([10.0, 20.0] | units.kg, y.previous_state().mass, 8)
        self.assertAlmostEqual([1.0, 2.0] | units.kg, y.previous_state().previous_state().mass, 8)
        self.assertEqual(y.previous_state().previous_state().previous_state(), None)

        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=False, overwrite_file=True, version="1.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5")
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertEqual(y.previous_state().previous_state(), None)

        os.remove("test_unit.hdf5")

    def test6b(self):
        print("Testing HDF5 io, with options")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="1.0")
        x.mass = [10.0, 20.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="1.0")
        x.mass = [100.0, 200.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="1.0", append_to_file=True)
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", copy_history=True, )
        y = y.previous_state()  # weirdness for version="1.0"
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual([10.0, 20.0] | units.kg, y.previous_state().mass, 8)
        self.assertAlmostEqual([1.0, 2.0] | units.kg, y.previous_state().previous_state().mass, 8)
        self.assertEqual(y.previous_state().previous_state().previous_state(), None)

        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=False, overwrite_file=True, version="1.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5")
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertEqual(y.previous_state().previous_state(), None)

        os.remove("test_unit.hdf5")

    def test6c(self):
        print("Testing HDF5 io, with options, version=2.0")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="2.0")
        x.mass = [10.0, 20.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="2.0")
        x.mass = [100.0, 200.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="2.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", close_file=False, copy_history=False)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual([10.0, 20.0] | units.kg, y.previous_state().mass, 8)
        self.assertAlmostEqual([1.0, 2.0] | units.kg, y.previous_state().previous_state().mass, 8)
        self.assertEqual(y.previous_state().previous_state().previous_state(), None)

        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=False, overwrite_file=True, version="2.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", copy_history=False)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertEqual(y.previous_state(), None)

        os.remove("test_unit.hdf5")

    def test6d(self):
        print("Testing HDF5 io, with options, version=2.0")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", version="2.0")
        x.mass = [10.0, 20.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="2.0")
        x.mass = [100.0, 200.0] | units.kg
        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=True, version="2.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", close_file=True, copy_history=True)
        y = y.previous_state()  # weirdness for version="1.0"
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual([10.0, 20.0] | units.kg, y.previous_state().mass, 8)
        self.assertAlmostEqual([1.0, 2.0] | units.kg, y.previous_state().previous_state().mass, 8)
        self.assertEqual(y.previous_state().previous_state().previous_state(), None)

        io.write_set_to_file(x, "test_unit.hdf5", "hdf5", append_to_file=False, overwrite_file=True, version="2.0")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5", copy_history=False)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertEqual(y.previous_state(), None)

        os.remove("test_unit.hdf5")

    def test7(self):
        print("Testing HDF5 io with a ParticlesSuperset")
        if os.path.exists("test_unit.hdf5"):
            os.remove("test_unit.hdf5")
        set1 = datamodel.Particles(2)
        set2 = datamodel.Particles(2)
        superset = datamodel.ParticlesSuperset([set1, set2])
        superset.mass = [1.0, 2.0, 3.0, 4.0] | units.kg
        superset.radius = [3.0, 4.0, 5.0, 6.0] | units.m
        superset.position = [[1, 2, 3], [3, 5, 6], [3, 2, 1], [-3, -5, -6]] | units.m
        superset.velocity = [[1, 2, 3], [3, 5, 6], [3, 2, 1], [-3, -5, -6]] | units.m / units.s
        io.write_set_to_file(superset, "test_unit.hdf5", "hdf5")
        y = io.read_set_from_file("test_unit.hdf5", "hdf5")

        self.assertAlmostEqual(superset.mass, y.mass, 8)
        self.assertAlmostEqual(superset.radius, y.radius, 8)
        self.assertAlmostEqual(superset.position, y.position, 8)
        self.assertAlmostEqual(superset.velocity, y.velocity, 8)

        os.remove("test_unit.hdf5")

    def test8(self):
        options = base.get_options_for_format('tsf')
        options.sort(key=lambda x: x[0])
        name, description, default = options[1]
        self.assertEqual(name, 'nbody_to_si_converter')
        self.assertEqual(description, 'NEMO datafiles store nbody data, provide a '
            'converter to store si data (None means no converter)')
        self.assertEqual(default, None)
        options = base.get_options_for_format('dyn')
        options.sort(key=lambda x: x[0])
        name, description, default = options[1]
        self.assertEqual(name, 'dynamics_mass_units')

        options = base.get_options_for_format('hdf5')
        options.sort(key=lambda x: x[0])
        name, description, default = options[1]
        self.assertEqual(name, 'append_to_file')
        self.assertTrue(description.find('If set to True, new data is appended to HDF5 files.') >= 0)
        self.assertTrue(description.find('If set to False, the existing file is removed and overwritten.') >= 0)
        self.assertEqual(default, False)

    def test9(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.kg
        self.assertRaises(AmuseException, io.write_set_to_file, x, "test_unit.bogus", "bogus",
            expected_message="You tried to load or save a file with fileformat 'bogus'"
                ", but this format is not in the supported formats list")

    def test10(self):
        print("Testing saving/loading timestamp in Starlab")
        x = datamodel.Particles(2)
        convert = nbody_system.nbody_to_si(1 | units.kg, 2 | units.m)
        x.mass = [1.0, 2.0] | units.kg
        x.position = [[1, 2, 3], [3, 5, 6]] | units.m
        x.velocity = [[1, 2, 3], [3, 5, 6]] | units.m / units.s
        current_time = 1.0 | units.Myr
        io.write_set_to_file(x.savepoint(current_time), "time_test_unit.dyn", "dyn", nbody_to_si_converter=convert)
        y = io.read_set_from_file("time_test_unit.dyn", "dyn", nbody_to_si_converter=convert)
        self.assertAlmostEqual(current_time, y.previous_state().get_timestamp(), 8, in_units=units.Myr)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | nbody_system.mass
        x.position = [[1, 2, 3], [3, 5, 6]] | nbody_system.length
        x.velocity = [[1, 2, 3], [3, 5, 6]] | nbody_system.speed
        current_time = 1.0 | nbody_system.time
        io.write_set_to_file(x.savepoint(current_time), "time_test_unit.dyn", "dyn")
        y = io.read_set_from_file("time_test_unit.dyn", "dyn")

        self.assertAlmostEqual(current_time, y.previous_state().get_timestamp(), 8, in_units=nbody_system.time)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        os.remove("time_test_unit.dyn")

    def test11(self):
        print("Testing saving/loading timestamp in NEMO")
        x = datamodel.Particles(2)
        convert = nbody_system.nbody_to_si(1 | units.kg, 2 | units.m)
        x.mass = [1.0, 2.0] | units.kg
        x.position = [[1, 2, 3], [3, 5, 6]] | units.m
        x.velocity = [[1, 2, 3], [3, 5, 6]] | units.m / units.s
        current_time = 1.0 | units.Myr
        io.write_set_to_file(x.savepoint(current_time), "time_test_unit.tsf", "tsf", nbody_to_si_converter=convert)
        y = io.read_set_from_file("time_test_unit.tsf", "tsf", nbody_to_si_converter=convert)

        self.assertAlmostEqual(current_time, y.previous_state().get_timestamp(), 8, in_units=units.Myr)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | nbody_system.mass
        x.position = [[1, 2, 3], [3, 5, 6]] | nbody_system.length
        x.velocity = [[1, 2, 3], [3, 5, 6]] | nbody_system.speed
        current_time = 1.0 | nbody_system.time
        io.write_set_to_file(x.savepoint(current_time), "time_test_unit.tsf", "tsf")
        y = io.read_set_from_file("time_test_unit.tsf", "tsf")

        self.assertAlmostEqual(current_time, y.previous_state().get_timestamp(), 8, in_units=nbody_system.time)
        self.assertAlmostEqual(x.mass, y.mass, 8)
        self.assertAlmostEqual(x.position, y.position, 8)
        self.assertAlmostEqual(x.velocity, y.velocity, 8)

        os.remove("time_test_unit.tsf")

    def test12(self):
        all_formats = sorted(base.registered_fileformat_processors.keys())
        for x in all_formats:
            options = base.get_options_for_format('txt')
            self.assertTrue(len(options) >= 0)

    def test_hdf5_compressed(self):
        options = {
            "version": "2.0",
            "compression": "gzip",
            "compression_opts": 5,
        }
        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            x = datamodel.Particles(2)
            x.mass = [1.0, 2.0] | units.kg
            io.write_set_to_file(x, filename, "hdf5", **options, overwrite_file=True)
            x.mass = [10.0, 20.0] | units.kg
            io.write_set_to_file(x, filename, "hdf5", append_to_file=True, **options)
            x.mass = [100.0, 200.0] | units.kg
            io.write_set_to_file(x, filename, "hdf5", append_to_file=True, **options)
            y = io.read_set_from_file(filename, "hdf5", copy_history=False, close_file=False)
            self.assertAlmostEqual(x.mass, y.mass, 8)
            self.assertAlmostEqual([10.0, 20.0] | units.kg, y.previous_state().mass, 8)
            self.assertAlmostEqual([1.0, 2.0] | units.kg, y.previous_state().previous_state().mass, 8)
            self.assertEqual(y.previous_state().previous_state().previous_state(), None)

            io.write_set_to_file(x, filename, "hdf5", append_to_file=False, overwrite_file=True, **options)
            y = io.read_set_from_file(filename, "hdf5")
            self.assertAlmostEqual(x.mass, y.mass, 8)
            self.assertEqual(y.previous_state().previous_state(), None)



import numpy
import time

from amuse.test import amusetest
from amuse.units import units
from amuse.datamodel.memory_storage import InMemoryGridAttributeStorage
from amuse.datamodel.memory_storage import InMemoryVectorQuantityAttribute


class TestInMemoryGridAttributeStorage(amusetest.TestCase):

    def test1(self):
        x = InMemoryGridAttributeStorage(5, 4, 3)
        i = (0, 1, 2, 3, 4)
        j = (1, 3, 1, 3, 1)
        k = (0, 2, 0, 2, 0)
        x.set_values_in_store(
            (i, j, k),
            ['a', 'b'],
            [2.0 | units.kg, 1.0 | units.m]
        )

        b, a = x.get_values_in_store(None, ['b', 'a'])
        print(b.shape, a.shape)
        self.assertEqual(b[0][1][0], 1.0 | units.m)
        self.assertEqual(b[0][0][0], 0.0 | units.m)
        self.assertEqual(a[0][1][0], 2.0 | units.kg)
        self.assertEqual(a[1][3][2], 2.0 | units.kg)
        self.assertEqual(a[1][2][2], 0.0 | units.kg)

        (b,) = x.get_values_in_store((numpy.s_[0:4], numpy.s_[1:4], numpy.s_[:]), ['a'])

        self.assertEqual(b[0][0][0], 2.0 | units.kg)
        self.assertEqual(b[0][0][2], 0.0 | units.kg)
        self.assertEqual(b[1][2][2], 2.0 | units.kg)
        self.assertEqual(b[2][0][0], 2.0 | units.kg)
        self.assertEqual(b[3][2][2], 2.0 | units.kg)

        self.assertEqual(b.sum(), 8.0 | units.kg)

        self.assertEqual(sorted(x.get_defined_attribute_names()), ["a", "b"])

    def test2(self):
        x = InMemoryGridAttributeStorage(5, 4, 3)
        i = (0, 1, 2, 3, 4)
        j = (1, 3, 1, 3, 1)
        k = (0, 2, 0, 2, 0)
        x.set_values_in_store(
            (i, j, k),
            ['a', 'b'],
            [2.0, 1.0]
        )

        b, a = x.get_values_in_store(None, ['b', 'a'])
        print(a)
        self.assertEqual(b[0][1][0], 1.0)
        self.assertEqual(b[0][0][0], 0.0)
        self.assertEqual(a[0][1][0], 2.0)
        self.assertEqual(a[1][3][2], 2.0)
        self.assertEqual(a[1][2][2], 0.0)

        (b,) = x.get_values_in_store((numpy.s_[0:4], numpy.s_[1:4], numpy.s_[:]), ['a'])

        self.assertEqual(b[0][0][0], 2.0)
        self.assertEqual(b[0][0][2], 0.0)
        self.assertEqual(b[1][2][2], 2.0)
        self.assertEqual(b[2][0][0], 2.0)
        self.assertEqual(b[3][2][2], 2.0)

        self.assertEqual(b.sum(), 8.0)

        self.assertEqual(sorted(x.get_defined_attribute_names()), ["a", "b"])

        y = x.copy()

        (b,) = y.get_values_in_store((numpy.s_[0:4], numpy.s_[1:4], numpy.s_[:]), ['a'])

        self.assertEqual(b[0][0][0], 2.0)
        self.assertEqual(b[0][0][2], 0.0)
        self.assertEqual(b[1][2][2], 2.0)

    def test3(self):
        x = InMemoryGridAttributeStorage(5, 4, 3)
        i = (0, 1, 2, 3, 4)
        j = (1, 3, 1, 3, 1)
        k = (0, 2, 0, 2, 0)
        x.set_values_in_store(
            (i, j, k),
            ['a', 'b'],
            [2.0 | units.kg, 1.0 | units.m]
        )

        b, a = x.get_values_in_store((0, 1, 0), ['b', 'a'])
        print(b, a)
        self.assertEqual(b, 1.0 | units.m)
        self.assertEqual(a, 2.0 | units.kg)
        b, a = x.get_values_in_store((0, 0, 0), ['b', 'a'])
        print(b, a)
        self.assertEqual(b, 0.0 | units.m)
        self.assertEqual(a, 0.0 | units.kg)


class TestInMemoryVectorQuantityAttribute(amusetest.TestCase):

    def test1(self):
        quantity = units.m.new_quantity(numpy.array([1.0, 2.0, 3.0]))
        attribute = InMemoryVectorQuantityAttribute('test', quantity.shape, quantity.unit)
        attribute.set_values(None, quantity)

        self.assertEqual(attribute.get_length(), 3)
        self.assertEqual(attribute.get_shape(), (3,))
        self.assertEqual(attribute.get_values([1, 2]), [2.0, 3.0] | units.m)

        attribute.increase_to_length(5)
        self.assertEqual(attribute.get_values(None), [1.0, 2.0, 3.0, 0.0, 0.0] | units.m)

    def test2(self):
        quantity = units.m.new_quantity(numpy.array([1.0, 2.0, 3.0]))
        attribute = InMemoryVectorQuantityAttribute('test', quantity.shape, quantity.unit)
        attribute.set_values(None, quantity)
        attribute.set_values([1, 2], [4.0, 5.0] | units.m)

        attribute.increase_to_length(5)
        self.assertEqual(attribute.get_values(None), [1.0, 4.0, 5.0, 0.0, 0.0] | units.m)

    def test3(self):
        quantity = units.m.new_quantity(numpy.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]))
        attribute = InMemoryVectorQuantityAttribute('test', quantity.shape, quantity.unit)
        attribute.set_values(None, quantity)
        self.assertEqual(attribute.get_values([1]),  [4.0, 5.0, 6.0] | units.m)
        self.assertEqual(attribute.get_shape(), (2, 3))
        attribute.increase_to_length(4)
        self.assertEqual(attribute.get_shape(), (4, 3))
        self.assertEqual(attribute.get_values(None), [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m)



from amuse.test.amusetest import TestWithMPI

import os
import sys

from amuse.community.hermite.interface import Hermite
from amuse.community.bhtree.interface import BHTree

import numpy
import threading
from amuse.units import nbody_system
from amuse.units import units
from amuse.datamodel import Particles
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestAmuseInterface(TestWithMPI):
    def new_system_sun_and_earth(self):
        result = Particles(2)
        sun = result[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = result[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))
        return result

    def evolve_model_unit_day(self, instance, particles, day):
        delta_days = 5
        for x in range(1, day + delta_days, delta_days):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(particles)
            particles.savepoint()

    def test1(self):
        from amuse.rfi import channel
        channel.MpiChannel.ensure_mpi_initialized()
        is_threaded = channel.MpiChannel.is_threaded()
        is_multithreading_supported = channel.MpiChannel.is_multithreading_supported()
        self.assertEqual(is_threaded, is_multithreading_supported)

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        bhtree = BHTree(convert_nbody)
        bhtree.initialize_code()
        bhtree.eps2_for_gravity = 0.001

        bhtree_particles = self.new_system_sun_and_earth()
        bhtree.particles.add_particles(bhtree_particles)

        if bhtree.legacy_interface.channel_type == 'mpi':
            from mpi4py import MPI
            if not MPI.Query_thread() == MPI.THREAD_MULTIPLE:
                bhtree.stop()
                self.skip("can only test parallel with multiple thread support in mpi implementation")

        hermite = Hermite(convert_nbody)
        hermite.dt_dia = 5000
        hermite.commit_parameters()

        hermite_particles = self.new_system_sun_and_earth()
        hermite.particles.add_particles(hermite_particles)

        thread1 = threading.Thread(target=self.evolve_model_unit_day, args=(bhtree, bhtree_particles, 10))
        thread2 = threading.Thread(target=self.evolve_model_unit_day, args=(hermite, hermite_particles, 10))

        thread1.start()
        thread2.start()

        thread1.join()
        thread2.join()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            earth = bhtree_particles[1]
            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")
            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            earth = hermite_particles[1]
            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")
            x_points_in_AU = [t_x2[1].value_in(units.AU) for t_x2 in x_points]
            y_points_in_AU = [t_x3[1].value_in(units.AU) for t_x3 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="g", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "parallel-earth-sun.svg")
            figure.savefig(output_file)

        bhtree.stop()
        hermite.stop()
        bhtree.stop()



import unittest
import numpy
from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.test import compile_tools
from amuse.support import exceptions

import os
import time
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi.tools import create_c
from amuse.rfi import channel
from amuse.rfi.core import *
codestring = """
#include <stdio.h>
#include <stdlib.h>

int echo_int(int int_in, int * int_out) {
    *int_out = int_in;
    if(int_in < 0) {
        return -1;
    } else {
        return 0;
    }
}
int echo_long_long_int(long long int int_in, long long int * int_out) {
    *int_out = int_in;
    if(int_in < 0) {
        return -1;
    } else {
        return 0;
    }
}

int echo_double(double in, double * out) {
    *out = in;
    return 0;
}

int echo_float(float in, float * out) {
    *out = in;
    return 0;
}
int echo_string(char * in, char ** out) {
    *out = in;
    return 0;
}

int echo_string_int(int inint, char * in, char ** out) {
    *out = in;
    return 0;
}

int echo_string_two(char * in1, char * in2, char ** out1, char ** out2) {
    *out1 = in1;
    *out2 = in2;
    return 0;
}

int print_string(char * in) {
    fprintf(stdout, "%s\\n", in);
    return 0;
}

int print_error_string(char * in) {
    fprintf(stderr, "%s\\n", in);
    return 0;
}

int echo_strings(char ** inout1, char ** inout2) {
    char * tmp;
    tmp = *inout1;
    *inout1 = *inout2;
    *inout2 = tmp;
    
    return 0;
}


void echo_array(int * in, int * out, int len) {
    int i = 0;
    for(i = 0; i < len; i++) {
        out[i] = in[i];
    }
}

int echo_array_with_result(int * in, int *out, int len) {
    int i = 0;
    for(i = 0; i < len; i++) {
        out[i] = in[i];
    }
    return -1;
}
        
int echo_2_int(int * int_in1, int * int_in2, int * int_out1, int * int_out2, int len) {
    int i = 0;
    for(i = 0; i < len; i++) {
        int_out1[i] = int_in1[i];
        int_out2[i] = int_in2[i];
    }
    
    return len;
}
int echo_3_int(int * i, int * j, int * k, int * l, int * m, int * int_out, int len) {
    int x = 0;
    for(x = 0; x < len; x++) {
        int_out[x] = i;
    }    
    return len;
}

int dummy_3_int(int i, int j, int k) {
    return 0;
}

int echo_inout_array_with_result(int * inout, int len) {
    int i = 0;
    for(i = 0; i < len; i++) {
        inout[i] = inout[i] + 10;
    }
    return 11;
}


int echo_logical(int in, int * out) {
    *out = in;
    return 0;
}
/*
int echo_string_array(char ** in, char ** out, int len) {
    int x = 0;
    for(x = 0; x < len; x++) {
        out[x] = in[x];
    }    
    return len;
}
*/
"""


class ForTestingInterface(CodeInterface):

    def __init__(self, exefile, **options):
        CodeInterface.__init__(self, exefile, **options)

    @remote_function(can_handle_array=True)
    def echo_int(int_in='int32'):
        returns(int_out='int32')

    @remote_function(can_handle_array=True)
    def echo_long_long_int(int_in='int64'):
        returns(int_out='int64')

    @remote_function(can_handle_array=True)
    def echo_double(double_in='float64'):
        returns(double_out='float64')

    @remote_function(can_handle_array=True)
    def echo_float(float_in='float32'):
        returns(float_out='float32')

    @remote_function(can_handle_array=True)
    def echo_string(string_in='string'):
        returns(string_out='string')

    @remote_function(can_handle_array=True)
    def echo_strings(string_inout1='string', string_inout2='string'):
        returns(string_inout1='string', string_inout2='string')

    @remote_function(can_handle_array=True)
    def echo_string_int(inint='int32', ins='echo'):
        returns(out='string')

    @remote_function(can_handle_array=True)
    def echo_string_two(in1='s', in2='echo'):
        returns(out1='s', out2='s')

    @remote_function(must_handle_array=True)
    def echo_array(len, int_in='int32'):
        returns(int_out='int32', __result=None)

    @remote_function(must_handle_array=True)
    def echo_array_with_result(len, int_in='int32'):
        returns(int_out='int32')

    # ~ #@legacy_function
    # ~ def return_string():
        # ~ function = LegacyFunctionSpecification()
        # ~ function.addParameter('string_in', dtype='string', direction=function.IN)
        # ~ function.result_type = 'string'
        # ~ function.can_handle_array = True
        # ~ return function

    @remote_function(must_handle_array=True)
    def echo_2_int(N, int_in1='int32', int_in2=numpy.int32(1)):
        returns(int_out1='int32', int_out2='int32')

    @remote_function(must_handle_array=True)
    def echo_3_int(i='int32', j='int32', k='int32', l=numpy.int32(0), m=numpy.int32(1)):
        returns(int_out='int32')

    @remote_function(must_handle_array=True)
    def echo_inout_array_with_result(in_out='int32'):
        returns(in_out='int32')

    @remote_function(can_handle_array=True)
    def echo_logical(input='bool'):
        returns(output='bool')

    @remote_function(can_handle_array=True)
    def print_string(string_in='string'):
        pass

    @remote_function
    def dummy_3_int(i='i', j='i', k='i'):
        pass

    @remote_function(can_handle_array=True)
    def print_error_string(string_in='string'):
        pass


class ForTesting(InCodeComponentImplementation):

    def __init__(self, exefile, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(exefile, **options), **options)

    def define_methods(self, object):
        object.add_method(
            'echo_int',
            (units.m,),
            (
                units.m,
                object.ERROR_CODE,
            )
        )


class TestCImplementationInterface(TestWithMPI):

    @classmethod
    def setup_class(cls):
        print("building...")
        cls.check_can_compile_modules()
        try:
            cls.exefile = compile_tools.build_worker(codestring, cls.get_path_to_results(),
                ForTestingInterface, write_header=False)
        except Exception as ex:
            print(ex)
            raise
        print("done")

    def test1(self):
        instance = ForTestingInterface(self.exefile)
        int_out, error = instance.echo_int(10)
        instance.stop()
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)

    def test2(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_double(4.0)
        instance.stop()
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test3(self):
        instance = ForTestingInterface(self.exefile)
        input = [1, 2, 3, 4]
        output, errors = instance.echo_int(input)
        instance.stop()
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test4(self):
        instance = ForTestingInterface(self.exefile)
        input = [1.0, 2.1, 3.3, 4.2]
        output, errors = instance.echo_double(input)
        instance.stop()
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test5(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_float(4.0)
        instance.stop()
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test6(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string("abc")
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out, "abc")

    def test7(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string(["abc", "def"])
        instance.stop()

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(out[0], "abc")
        self.assertEqual(out[1], "def")

    def test8(self):
        instance = ForTestingInterface(self.exefile)
        out1, out2, error = instance.echo_strings("abc", "def")
        instance.stop()

        self.assertEqual(error, 0)
        self.assertEqual(out1, "def")
        self.assertEqual(out2, "abc")

    def test9(self):
        instance = ForTestingInterface(self.exefile)
        str1_out, str2_out, error = instance.echo_strings(["abc", "def"], ["ghi", "jkl"])
        instance.stop()

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(str1_out[0], "ghi")
        self.assertEqual(str1_out[1], "jkl")
        self.assertEqual(str2_out[0], "abc")
        self.assertEqual(str2_out[1], "def")

    def xtest10(self):
        """test for ticket #74, 'running out of os file descriptors'


        Note: this test takes a very long time, to enable it
        remove the 'X' infront of the test name, to disable it
        add an 'X'.
        Also note: to test this, you best set the ulimit
        to a low number (but not too low), for example
        ulimit -n 400
        """
        for x in range(400):
            instance = ForTestingInterface(self.exefile)
            out, error = instance.echo_float(4.0)
            if x % 100 == 0:
                print("x:", x)
            instance.stop()

    def test11(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints,) = instance.echo_array([4, 5, 6])
        instance.stop()
        print(output_ints)
        self.assertEqual(output_ints[0], 4)
        self.assertEqual(output_ints[1], 5)
        self.assertEqual(output_ints[2], 6)

    def test12(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints, error) = instance.echo_array_with_result([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 4)
        self.assertEqual(output_ints[1], 5)
        self.assertEqual(output_ints[2], 6)

        self.assertEqual(error[0], -1)
        self.assertEqual(error[1], -1)
        self.assertEqual(error[2], -1)

    def test13(self):
        instance = ForTesting(self.exefile)
        self.assertRaises(exceptions.AmuseException, instance.echo_int, [-1, -2] | units.m,
            expected_message="Error when calling 'echo_int' of a 'ForTesting', errorcode is -1")
        instance.stop()

    def test14(self):
        instance = ForTesting(self.exefile)
        self.assertRaises(exceptions.CodeException, lambda: instance.echo_int())
        old_id = instance.legacy_interface.echo_int.specification.id
        instance.legacy_interface.echo_int.specification.id = -9
        self.assertRaises(exceptions.CodeException, lambda: instance.echo_int(1 | units.m))
        instance.legacy_interface.echo_int.specification.id = old_id
        instance.echo_int(1 | units.m)
        instance.stop()

    def test15(self):
        instance = ForTesting(self.exefile)
        output_ints1, output_ints2 = instance.echo_2_int([1, 2], [3, 4])
        output_ints3, output_ints4 = instance.echo_2_int([1, 2, 3])
        output_ints5, output_ints6 = instance.echo_2_int([5], [0])
        output_ints7, output_ints8 = instance.echo_2_int([5])
        instance.stop()
        self.assertEqual(output_ints1[1], 2)
        self.assertEqual(output_ints2[0], 3)
        self.assertEqual(output_ints2[1], 4)
        for i in range(3):
            self.assertEqual(output_ints3[i], i + 1)
            self.assertEqual(output_ints4[i], 1)
        self.assertEqual(output_ints5[0], 5)
        self.assertEqual(output_ints6[0], 0)
        self.assertEqual(output_ints7[0], 5)
        self.assertEqual(output_ints8[0], 1)

    def test16(self):
        instance = ForTesting(self.exefile)

        self.assertRaises(exceptions.AmuseException, lambda: instance.echo_int([]))
        instance.stop()

    def test17(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints, error) = instance.echo_inout_array_with_result([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 14)
        self.assertEqual(output_ints[1], 15)
        self.assertEqual(output_ints[2], 16)

        self.assertEqual(error[0], 11)
        self.assertEqual(error[1], 11)
        self.assertEqual(error[2], 11)

    def test18(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_logical([True, False, True])
        instance.stop()
        self.assertEqual(out, [True, False, True])
        self.assertEqual(error, 0)

    def test19(self):
        instance = ForTestingInterface(self.exefile)
        print(3935559000370003845)
        int_out, error = instance.echo_long_long_int(3935559000370003845)
        instance.stop()
        self.assertEqual(int_out, 3935559000370003845)
        self.assertEqual(error, 0)

    def xtest20(self):

        #
        # TURNED OFF support for redirection,
        # by default output is redirected to /dev/null
        # if you need file, use the support from your mpi implementation
        #
        if os.path.exists("pout.000"):
            os.remove("pout.000")
        if os.path.exists("perr.000"):
            os.remove("perr.000")

        x = ForTesting(self.exefile, redirect_stderr_file='perr', redirect_stdout_file='pout', redirection="file")
        x.print_string("abc")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc")

        self.assertTrue(os.path.exists("perr.000"))
        with open("perr.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "exex")

        x = ForTesting(self.exefile, redirect_stderr_file='pout', redirect_stdout_file='pout', redirection="file")
        x.print_string("def")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc\ndef\nexex")

    def test21(self):
        instance = ForTestingInterface(self.exefile)
        (output1, error1) = instance.internal__get_message_polling_interval()
        error2 = instance.internal__set_message_polling_interval(1234)
        (output2, error3) = instance.internal__get_message_polling_interval()
        instance.internal__set_message_polling_interval(0)
        instance.stop()
        self.assertEqual(error1, 0)
        self.assertEqual(output1, 0)
        self.assertEqual(error2, 0)
        self.assertEqual(error3, 0)
        self.assertEqual(output2, 1234)

    def test22(self):
        self.check_for_mpi()
        instance = ForTestingInterface(self.exefile)
        t0 = time.time()
        (output1, error1) = instance.internal__get_message_polling_interval()
        t1 = time.time()
        error2 = instance.internal__set_message_polling_interval(500 * 1000)
        t2 = time.time()
        (output2, error3) = instance.internal__get_message_polling_interval()
        t3 = time.time()
        instance.stop()
        self.assertEqual(error1, 0)
        self.assertEqual(output1, 0)
        self.assertEqual(error2, 0)
        self.assertEqual(error3, 0)
        self.assertEqual(output2, 500 * 1000)
        # ~ print t1 - t0, t3 - t2
        # ~ self.assertTrue((t3 - t2) > 0.25)

    def test23(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string_int(1)
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out, "echo")

    def test24(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string_int(1, "abc")
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out, "abc")

    def test25(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string_int([1, 2])
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out[0], "echo")
        self.assertEqual(out[1], "echo")

    def test26(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string_int([1, 2], ["abc", "def"])
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out[0], "abc")
        self.assertEqual(out[1], "def")

    def test27(self):

        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string_int([1, 2], "abc")
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out[0], "abc")
        self.assertEqual(out[1], "abc")

    def test28(self):
        instance = ForTestingInterface(self.exefile)
        out1, out2, error = instance.echo_string_two(["one", "two"], "three")
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out1[0], "one")
        self.assertEqual(out1[1], "two")
        self.assertEqual(out2[0], "three")
        self.assertEqual(out2[1], "three")

    def test29(self):
        self.check_for_mpi()
        instance1 = ForTestingInterface(self.exefile)
        instance2 = ForTestingInterface(self.exefile)
        portname, error = instance1.internal__open_port()
        self.assertTrue(len(portname) > 0)
        request1 = instance1.internal__accept_on_port.asynchronous(portname)
        request2 = instance2.internal__connect_to_port.asynchronous(portname)
        request1.wait()
        request2.wait()
        port_id1, error1 = request1.result()
        port_id2, error2 = request2.result()
        self.assertTrue(port_id1 >= 0)
        self.assertTrue(port_id2 >= 0)
        self.assertEqual(error1, 0)
        self.assertEqual(error2, 0)

    def test30(self):
        from amuse.support.interface import ConvertArgumentsException
        instance = ForTesting(self.exefile)
        self.assertRaises(ConvertArgumentsException, instance.dummy_3_int, 2, 3, i=1, expected_message="got multiple values for argument 'i' of method dummy_3_int")
        instance.stop()

    @unittest.skip
    def test31(self):
        import time
        instance = ForTestingInterface(self.exefile)
        N = 5000
        t1 = time.time()
        for i in range(N):
            res, err = instance.echo_int([i])
        t2 = time.time()
        print("1 time:", t2-t1, (t2-t1)/N)
        instance.stop()

        instance = ForTesting(self.exefile)
        N = 5000
        t1 = time.time()
        for i in range(N):
            res = instance.echo_int([i] | units.m)
        t2 = time.time()
        print("2 time:", t2-t1, (t2-t1)/N)
        instance.stop()



from amuse.test.amusetest import TestWithMPI

from amuse.support.exceptions import AmuseException
from amuse.ext.halogen_model import new_halogen_model
from amuse.units import nbody_system, units


class NewHalogenModelTests(TestWithMPI):

    def test1(self):
        number_of_particles = 100
        particles = new_halogen_model(number_of_particles, alpha=2.0, beta=5.0, gamma=0.0, random_seed=1.0)

        self.assertEqual(len(particles), number_of_particles)
        self.assertAlmostEqual(particles.total_mass(), 1.0 | nbody_system.mass)
        self.assertAlmostEqual(particles.kinetic_energy(),
            0.17345836639 | nbody_system.energy)  # for number_of_particles = 100

        self.assertRaises(AmuseException, new_halogen_model, number_of_particles, expected_message="Error when calling 'commit_parameters' of a 'Halogen', errorcode is -2, error is "
            "'Missing or bad parameter for halo (see amuse/community/halogen/src/doc for details on required parameters).'")

    def test2(self):
        number_of_particles = 1000
        black_hole_mass = 1.0e6 | units.MSun
        stellar_mass = number_of_particles | units.MSun
        scale_radius = 0.1 | units.parsec

        converter = nbody_system.nbody_to_si(black_hole_mass, scale_radius)

        particles = new_halogen_model(number_of_particles, convert_nbody=converter,
            alpha=1.0, beta=3.0, gamma=1.0,  # NFW
            random_seed=1.0, redirection='none',
            black_hole_mass=black_hole_mass, total_mass=stellar_mass,
            cutoff_radius=10.0*scale_radius, scale_radius=scale_radius)

        self.assertEqual(len(particles), number_of_particles + 1)
        self.assertAlmostEqual(particles[-1].mass, black_hole_mass)
        self.assertAlmostEqual(particles.total_mass(), black_hole_mass + stellar_mass)
        self.assertAlmostRelativeEquals(particles.kinetic_energy(),
            2.27538127277e+43 | units.J, 10)



import numpy
from amuse.test.amusetest import TestCase
from amuse.units import units
from amuse.datamodel import Particles
from amuse.ext.particles_with_color import *


class TestParticlesWithColor(TestCase):

    def test1(self):
        print("Test new_particles_with_color")
        with_color = new_particles_with_color(
            Particles(mass=[0.1, 1, 2, 3] | units.MSun),
            lambda mass: numpy.select([mass < 2.5 | units.MSun], [1]),
            lambda mass: numpy.select([numpy.logical_and(0.5 | units.MSun < mass, mass < 2.5 | units.MSun)], [1]),
            lambda mass: numpy.select([mass > 1.5 | units.MSun], [1])
        )
        self.assertEqual(with_color.red,   [1, 1, 1, 0])
        self.assertEqual(with_color.green, [0, 1, 1, 0])
        self.assertEqual(with_color.blue,  [0, 0, 1, 1])
        self.assertEqual(with_color.color, [[1, 0, 0], [1, 1, 0], [1, 1, 1], [0, 0, 1]])
        with_color.mass = [3, 2, 1, 0.1] | units.MSun
        self.assertEqual(with_color.red,   [0, 1, 1, 1])
        self.assertEqual(with_color.green, [0, 1, 1, 0])
        self.assertEqual(with_color.blue,  [1, 1, 0, 0])
        self.assertEqual(with_color.color, [[0, 0, 1], [1, 1, 1], [1, 1, 0], [1, 0, 0]])

    def test2(self):
        print("Test new_particles_with_blackbody_color from temperature")
        original = Particles(temperature=[100, 1000, 2000, 5000, 10000, 40000] | units.K)
        with_color = new_particles_with_blackbody_color(original)
        self.assertAlmostEqual(with_color.red,   [1.0000, 1.0000, 1.0000, 1.0000, 0.6324, 0.3565], 3)
        self.assertAlmostEqual(with_color.green, [0.0337, 0.0337, 0.2434, 0.7872, 0.7081, 0.4747], 3)
        self.assertAlmostEqual(with_color.blue,  [0.0000, 0.0000, 0.0000, 0.5797, 1.0000, 1.0000], 3)
        self.assertEqual(with_color.color.shape, (6, 3))

    def test3(self):
        print("Test new_particles_with_blackbody_color from internal energy")
        original = Particles(u=u_from_T([100, 1000, 2000, 5000, 10000, 40000] | units.K))
        with_color = new_particles_with_blackbody_color(original)
        self.assertAlmostEqual(with_color.red,   [1.0000, 1.0000, 1.0000, 1.0000, 0.6324, 0.3565], 3)
        self.assertAlmostEqual(with_color.green, [0.0337, 0.0337, 0.2434, 0.7872, 0.7081, 0.4747], 3)
        self.assertAlmostEqual(with_color.blue,  [0.0000, 0.0000, 0.0000, 0.5797, 1.0000, 1.0000], 3)
        self.assertEqual(with_color.color.shape, (6, 3))
